"""Backend agnostic functions for solving matrices either fully or partially.
"""
import functools
import warnings

import numpy as np
import scipy.linalg as sla
import scipy.sparse.linalg as spla

from ..utils import raise_cant_find_library_function
from ..core import qarray, dag, issparse, isdense, vdot, ldmul
from .numpy_linalg import (
    eig_numpy,
    eigs_numpy,
    svds_numpy,
)

from .scipy_linalg import (
    eigs_scipy,
    eigs_lobpcg,
    svds_scipy,
)
from . import SLEPC4PY_FOUND

if SLEPC4PY_FOUND:
    from .mpi_launcher import (
        eigs_slepc_spawn,
        mfn_multiply_slepc_spawn,
        svds_slepc_spawn,
    )
    from .slepc_linalg import eigs_slepc, svds_slepc, mfn_multiply_slepc
else:  # pragma: no cover
    eigs_slepc = raise_cant_find_library_function("slepc4py")
    eigs_slepc_spawn = raise_cant_find_library_function("slepc4py")
    svds_slepc = raise_cant_find_library_function("slepc4py")
    svds_slepc_spawn = raise_cant_find_library_function("slepc4py")
    mfn_multiply_slepc = raise_cant_find_library_function("slepc4py")
    mfn_multiply_slepc_spawn = raise_cant_find_library_function("slepc4py")


# --------------------------------------------------------------------------- #
#                          Partial eigendecomposition                         #
# --------------------------------------------------------------------------- #

def choose_backend(A, k, int_eps=False, B=None):
    """Pick a backend automatically for partial decompositions.
    """
    # LinOps -> not possible to simply convert to dense or use MPI processes
    A_is_linop = isinstance(A, spla.LinearOperator)
    B_is_linop = isinstance(B, spla.LinearOperator)

    # small array or large part of subspace requested
    small_d_big_k = A.shape[0] ** 2 / k < (10000 if int_eps else 2000)

    if small_d_big_k and not (A_is_linop or B_is_linop):
        return "NUMPY"

    # slepc seems faster for sparse, dense and LinearOperators
    if SLEPC4PY_FOUND and not B_is_linop:

        # only spool up an mpi pool for big sparse matrices though
        if issparse(A) and A.nnz > 10000:
            return 'SLEPC'

        return 'SLEPC-NOMPI'

    return 'SCIPY'


_EIGS_METHODS = {
    'NUMPY': eigs_numpy,
    'SCIPY': eigs_scipy,
    'LOBPCG': eigs_lobpcg,
    'SLEPC': eigs_slepc_spawn,
    'SLEPC-NOMPI': eigs_slepc,
}


def eigensystem_partial(A, k, isherm, *, B=None, which=None, return_vecs=True,
                        sigma=None, ncv=None, tol=None, v0=None, sort=True,
                        backend=None, fallback_to_scipy=False, **backend_opts):
    """Return a few eigenpairs from an operator.

    Parameters
    ----------
    A : sparse, dense or linear operator
        The operator to solve for.
    k : int
        Number of eigenpairs to return.
    isherm : bool
        Whether to use hermitian solve or not.
    B : sparse, dense or linear operator, optional
        If given, the RHS operator defining a generalized eigen problem.
    which : {'SA', 'LA', 'LM', 'SM', 'TR'}
        Where in spectrum to take eigenvalues from (see
        :func:``scipy.sparse.linalg.eigsh``)
    return_vecs : bool, optional
        Whether to return the eigenvectors.
    sigma : float, optional
        Which part of spectrum to target, implies which='TR' if which is None.
    ncv : int, optional
        number of lanczos vectors, can use to optimise speed
    tol : None or float
        Tolerance with which to find eigenvalues.
    v0 : None or 1D-array like
        An initial vector guess to iterate with.
    sort : bool, optional
        Whether to explicitly sort by ascending eigenvalue order.
    backend : {'AUTO', 'NUMPY', 'SCIPY',
               'LOBPCG', 'SLEPC', 'SLEPC-NOMPI'}, optional
        Which solver to use.
    fallback_to_scipy : bool, optional
        If an error occurs and scipy is not being used, try using scipy.
    backend_opts
        Supplied to the backend solver.

    Returns
    -------
    elk : (k,) array
        The ``k`` eigenvalues.
    evk : (d, k) array
        Array with ``k`` eigenvectors as columns if ``return_vecs``.
    """
    settings = {
        'k': k,
        'B': B,
        'which': ("SA" if (which is None) and (sigma is None) else
                  "TR" if (which is None) and (sigma is not None) else
                  which),
        'return_vecs': return_vecs,
        'sigma': sigma,
        'isherm': isherm,
        'ncv': ncv,
        'sort': sort,
        'tol': tol,
        'v0': v0,
    }

    # Choose backend to perform the decompostion
    bkd = 'AUTO' if backend is None else backend.upper()
    if bkd == 'AUTO':
        bkd = choose_backend(A, k, sigma is not None, B=B)

    try:
        return _EIGS_METHODS[bkd](A, **settings, **backend_opts)

    # sometimes e.g. lobpcg fails, worth trying scipy
    except Exception as e:  # pragma: no cover
        if fallback_to_scipy and (bkd != 'SCIPY'):
            warnings.warn(
                f"`eigensystem_partial` with backend '{bkd}' failed, trying "
                "again with scipy. Set ``fallback_to_scipy=False`` to avoid "
                "this and see the full error.")

            return eigs_scipy(A, **settings, **backend_opts)
        else:
            raise e


# --------------------------------------------------------------------------- #
#                        Full eigendecomposition                              #
# --------------------------------------------------------------------------- #

def eigensystem(A, isherm, *, k=-1, sort=True, return_vecs=True, **kwargs):
    """Find all or some eigenpairs of an operator.

    Parameters
    ----------
    A : operator
        The operator to decompose.
    isherm : bool
        Whether the operator is assumed to be hermitian or not.
    k : int, optional
        If negative, find all eigenpairs, else perform partial
        eigendecomposition and find ``k`` pairs. See
        :func:`~quimb.linalg.base_linalg.eigensystem_partial`.
    sort : bool, optional
        Whether to sort the eigenpairs in ascending eigenvalue order.
    kwargs
        Supplied to the backend function.

    Returns
    -------
    el : (k,) array
        Eigenvalues.
    ev : (d, k) array
        Corresponding eigenvectors as columns of array, such that
        ``ev @ diag(el) @ ev.H == A``.
    """
    if k < 0:
        return eig_numpy(A, isherm=isherm, sort=sort,
                         return_vecs=return_vecs, **kwargs)

    return eigensystem_partial(A, k=k, isherm=isherm, sort=sort,
                               return_vecs=return_vecs, **kwargs)


eig = functools.partial(eigensystem, isherm=False, return_vecs=True)
eigh = functools.partial(eigensystem, isherm=True, return_vecs=True)
eigvals = functools.partial(eigensystem, isherm=False, return_vecs=False)
eigvalsh = functools.partial(eigensystem, isherm=True, return_vecs=False)


@functools.wraps(eigensystem)
def eigenvectors(A, isherm, *, sort=True, **kwargs):
    return eigensystem(A, isherm=isherm, sort=sort, **kwargs)[1]


eigvecs = functools.partial(eigenvectors, isherm=False)
eigvecsh = functools.partial(eigenvectors, isherm=True)


def groundstate(ham, **kwargs):
    """Alias for finding lowest eigenvector only.
    """
    return eigvecsh(ham, k=1, which='SA', **kwargs)


def groundenergy(ham, **kwargs):
    """Alias for finding lowest eigenvalue only.
    """
    return eigvalsh(ham, k=1, which='SA', **kwargs)[0]


def bound_spectrum(A, backend='auto', **kwargs):
    """Return the smallest and largest eigenvalue of hermitian operator ``A``.
    """
    el_min = eigvalsh(A, k=1, which='SA', backend=backend, **kwargs)[0]
    el_max = eigvalsh(A, k=1, which='LA', backend=backend, **kwargs)[0]
    return el_min, el_max


def _rel_window_to_abs_window(el_min, el_max, w_0, w_sz=None):
    """Convert min/max eigenvalues and relative window to absolute values.

    Parameters
    ----------
    el_min : float
        Smallest eigenvalue.
    el_max : float
        Largest eigenvalue.
    w_0 : float [0.0 - 1.0]
        Relative window centre.
    w_sz : float, optional
        Relative window width.

    Returns
    -------
    l_0[, l_min, l_max]:
        Absolute value of centre of window, lower and upper intervals if a
        window size is specified.
    """
    el_range = el_max - el_min
    el_w_0 = el_min + w_0 * el_range
    if w_sz is not None:
        el_w_min = el_w_0 - w_sz * el_range / 2
        el_w_max = el_w_0 + w_sz * el_range / 2
        return el_w_0, el_w_min, el_w_max
    return el_w_0


def eigh_window(A, w_0, k, w_sz=None, backend='AUTO',
                return_vecs=True, offset_const=1 / 104729, **kwargs):
    """ Return mid-spectrum eigenpairs from a hermitian operator.

    Parameters
    ----------
    A : (d, d) operator
        Operator to retrieve eigenpairs from.
    w_0 : float [0.0, 1.0]
        Relative window centre to retrieve eigenpairs from.
    k : int
        Target number of eigenpairs to retrieve.
    w_sz : float, optional
        Relative maximum window width within which to keep eigenpairs.
    backend : str, optional
        Which :func:`~quimb.eigh` backend to use.
    return_vecs : bool, optional
        Whether to return eigenvectors as well.
    offset_const : float, optional
        Small fudge factor (relative to window range) to avoid 1 / 0 issues.

    Returns
    -------
    el : (k,) array
        Eigenvalues around w_0.
    ev : (d, k) array
        The eigenvectors, if ``return_vecs=True``.
    """
    w_sz = w_sz if w_sz is not None else 1.1

    if isdense(A) or backend.upper() == 'NUMPY':
        if return_vecs:
            lk, vk = eigh(A.A if issparse(A) else A, **kwargs)
        else:
            lk = eigvalsh(A.A if issparse(A) else A, **kwargs)

        lmin, lmax = lk[0], lk[-1]
        l_w0, l_wmin, l_wmax = _rel_window_to_abs_window(lmin, lmax, w_0, w_sz)

    else:
        lmin, lmax = bound_spectrum(A, backend=backend, **kwargs)
        l_w0, l_wmin, l_wmax = _rel_window_to_abs_window(lmin, lmax, w_0, w_sz)
        l_w0 += (lmax - lmin) * offset_const  # for 1/0 issues

        if return_vecs:
            lk, vk = eigh(A, k=k, sigma=l_w0, backend=backend, **kwargs)
        else:
            lk = eigvalsh(A, k=k, sigma=l_w0, backend=backend, **kwargs)

    # Trim eigenpairs from beyond window
    in_window = (lk > l_wmin) & (lk < l_wmax)

    if return_vecs:
        return lk[in_window], vk[:, in_window]

    return lk[in_window]


def eigvalsh_window(*args, **kwargs):
    """Alias for only finding the eigenvalues in a relative window.
    """
    return eigh_window(*args, return_vecs=False, **kwargs)


def eigvecsh_window(*args, **kwargs):
    """Alias for only finding the eigenvectors in a relative window.
    """
    return eigh_window(*args, return_vecs=True, **kwargs)[1]


# -------------------------------------------------------------------------- #
# Partial singular value decomposition                                       #
# -------------------------------------------------------------------------- #

def svd(A, return_vecs=True):
    """Compute full singular value decomposition of an operator, using numpy.

    Parameters
    ----------
    A : (m, n) array
        The operator.
    return_vecs : bool, optional
        Whether to return the singular vectors.

    Returns
    -------
    U : (m, k) array
        Left singular vectors (if ``return_vecs=True``) as columns.
    s : (k,) array
        Singular values.
    VH : (k, n) array
        Right singular vectors (if ``return_vecs=True``) as rows.
    """
    try:
        return np.linalg.svd(A, full_matrices=False, compute_uv=return_vecs)

    except np.linalg.linalg.LinAlgError:  # pragma: no cover
        warnings.warn("Numpy SVD failed, trying again with different driver.")
        return sla.svd(A, full_matrices=False, compute_uv=return_vecs,
                       lapack_driver='gesvd')


_SVDS_METHODS = {
    'SLEPC': svds_slepc_spawn,
    'SLEPC-NOMPI': svds_slepc,
    'NUMPY': svds_numpy,
    'SCIPY': svds_scipy,
}


def svds(A, k, ncv=None, return_vecs=True, backend='AUTO', **kwargs):
    """Compute the partial singular value decomposition of an operator.

    Parameters
    ----------
    A : dense, sparse or linear operator
        The operator to decompose.
    k : int, optional
        number of singular value (triplets) to retrieve
    ncv : int, optional
        Number of lanczos vectors to use performing decomposition.
    return_vecs : bool, optional
        Whether to return the left and right vectors
    backend : {'AUTO', 'SCIPY', 'SLEPC', 'SLEPC-NOMPI', 'NUMPY'}, optional
        Which solver to use to perform decomposition.

    Returns
    -------
    (Uk,) sk (, VHk) :
        Singular value(s) (and vectors) such that ``Uk @ np.diag(sk) @ VHk``
        approximates ``A``.
    """
    settings = {
        'k': k,
        'ncv': ncv,
        'return_vecs': return_vecs}

    bkd = (choose_backend(A, k, False) if backend in {'auto', 'AUTO'} else
           backend.upper())
    svds_func = _SVDS_METHODS[bkd.upper()]

    return svds_func(A, **settings, **kwargs)


# -------------------------------------------------------------------------- #
# Norms and other quantities based on decompositions                         #
# -------------------------------------------------------------------------- #

def norm_2(A, **kwargs):
    """Return the 2-norm of operator, ``A``, i.e. the largest singular value.
    """
    return svds(A, k=1, return_vecs=False, **kwargs)[0]


def norm_fro_dense(A):
    """Frobenius norm for dense matrices
    """
    return vdot(A, A).real**0.5


def norm_fro_sparse(A):
    return vdot(A.data, A.data).real**0.5


def norm_trace_dense(A, isherm=False):
    """Returns the trace norm of operator ``A``, that is,
    the sum of the absolute eigenvalues.
    """
    if isherm:
        return abs(eigvalsh(A)).sum()
    else:
        return svd(A, return_vecs=False).sum()


def norm(A, ntype=2, **kwargs):
    """Operator norms.

    Parameters
    ----------
    A : operator
        The operator to find norm of.
    ntype : str
        Norm to calculate, if any of:

        - {2, '2', 'spectral'}: largest singular value
        - {'f', 'fro'}: frobenius norm
        - {'t', 'nuc', 'tr', 'trace'}: sum of singular values

    Returns
    -------
    x : float
        The operator norm.
    """
    types = {'2': '2', 2: '2', 'spectral': '2',
             'f': 'f', 'fro': 'f',
             't': 't', 'trace': 't', 'nuc': 't', 'tr': 't'}
    methods = {('2', 0): norm_2,
               ('2', 1): norm_2,
               ('t', 0): norm_trace_dense,
               ('f', 0): norm_fro_dense,
               ('f', 1): norm_fro_sparse}
    return methods[(types[ntype], issparse(A))](A, **kwargs)


# --------------------------------------------------------------------------- #
#                               Matrix functions                              #
# --------------------------------------------------------------------------- #

def expm(A, herm=False):
    """Matrix exponential, can be accelerated if explicitly hermitian.

    Parameters
    ----------
    A : dense or sparse operator
        Operator to exponentiate.
    herm : bool, optional
        If True (not default), and ``A`` is dense, digonalize the matrix
        in order to perform the exponential.
    """
    if issparse(A):
        # convert to and from csc to suppress scipy warning
        return spla.expm(A.tocsc()).tocsr()
    elif not herm:
        return qarray(spla.expm(A))
    else:
        evals, evecs = eigh(A)
        return evecs @ ldmul(np.exp(evals), dag(evecs))


_EXPM_MULTIPLY_METHODS = {
    'SCIPY': spla.expm_multiply,
    'SLEPC': functools.partial(mfn_multiply_slepc_spawn, fntype='exp'),
    'SLEPC-KRYLOV': functools.partial(
        mfn_multiply_slepc_spawn, fntype='exp', MFNType='KRYLOV'),
    'SLEPC-EXPOKIT': functools.partial(
        mfn_multiply_slepc_spawn, fntype='exp', MFNType='EXPOKIT'),
    'SLEPC-NOMPI': functools.partial(mfn_multiply_slepc, fntype='exp'),
}


def expm_multiply(mat, vec, backend="AUTO", **kwargs):
    """Compute the action of ``expm(mat)`` on ``vec``.

    Parameters
    ----------
    mat : operator
        Operator with which to act with exponential on ``vec``.
    vec : vector-like
        Vector to act with exponential of operator on.
    backend : {'AUTO', 'SCIPY', 'SLEPC', 'SLEPC-KRYLOV', 'SLEPC-EXPOKIT'}
        Which backend to use.
    kwargs
        Supplied to backend function.

    Returns
    -------
    vector
        Result of ``expm(mat) @ vec``.
    """
    if backend == 'AUTO':
        if SLEPC4PY_FOUND and vec.size > 2**10:
            backend = 'SLEPC'
        else:
            backend = 'SCIPY'

    return _EXPM_MULTIPLY_METHODS[backend.upper()](mat, vec, **kwargs)


def sqrtm(A, herm=True):
    """Matrix square root, can be accelerated if explicitly hermitian.

    Parameters
    ----------
    A : dense array
        Operator to take square root of.
    herm : bool, optional
        If True (the default), and ``A`` is dense, digonalize the matrix
        in order to take the square root.

    Returns
    -------
    array
    """
    if issparse(A):
        raise NotImplementedError("No sparse sqrtm available.")
    elif not herm:
        return qarray(sla.sqrtm(A))
    else:
        evals, evecs = eigh(A)
        return evecs @ ldmul(np.sqrt(evals.astype(complex)), dag(evecs))


class IdentityLinearOperator(spla.LinearOperator):
    """Get a ``LinearOperator`` representation of the identity operator,
    scaled by ``factor``.

    Parameters
    ----------
    size : int
        The size of the identity.
    factor : float
        The coefficient of the identity.

    Examples
    --------

    >>> I3 = IdentityLinearOperator(100, 1/3)
    >>> p = rand_ket(100)
    >>> np.allclose(I3 @ p, p / 3)
    True
    """

    def __init__(self, size, factor=1):
        self.factor = factor
        super().__init__(dtype=np.array(factor).dtype, shape=(size, size))

    def _matvec(self, vec):
        return self.factor * vec

    def _rmatvec(self, vec):
        return self.factor * vec

    def _matmat(self, mat):
        return self.factor * mat


class Lazy:
    """A simple class representing an unconstructed matrix. This can be passed
    to, for example, MPI workers, who can then construct the matrix themselves.
    The main function ``fn`` should ideally take an ``ownership`` keyword to
    avoid forming every row.

    This is essentially like using ``functools.partial`` and assigning the
    ``shape`` attribute.

    Parameters
    ----------
    fn : callable
        A function that constructs an operator.
    shape :
        Shape of the constructed operator.
    args
        Supplied to ``fn``.
    kwargs
        Supplied to ``fn``.

    Returns
    -------
    Lazy : callable

    Examples
    --------
    Setup the lazy operator:

    >>> H_lazy = Lazy(ham_heis, n=10, shape=(2**10, 2**10), sparse=True)
    >>> H_lazy
    <Lazy(ham_heis, shape=(1024, 1024), dtype=None)>

    Build a matrix slice (usually done automatically by e.g. ``eigs``):

    >>> H_lazy(ownership=(256, 512))
    <256x1024 sparse matrix of type '<class 'numpy.float64'>'
            with 1664 stored elements in Compressed Sparse Row format>
    """

    def __init__(self, fn, *args, shape=None, factor=None, **kwargs):
        if shape is None:
            raise TypeError("`shape` must be specified.")
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.shape = shape
        self.factor = factor
        self.dtype = None

    def __imul__(self, x):
        if self.factor is None:
            self.factor = x
        else:
            self.factor = self.factor * x

    def __mul__(self, x):
        if self.factor is not None:
            x = x * self.factor
        return Lazy(self.fn, *self.args, shape=self.shape,
                    factor=x, **self.kwargs)

    def __rmul__(self, x):
        return self.__mul__(x)

    def __call__(self, **kwargs):
        A = self.fn(*self.args, **self.kwargs, **kwargs)

        # check if any prefactors have been set
        if self.factor is not None:
            # try inplace first
            try:
                A *= self.factor
            except (ValueError, TypeError):
                A = self.factor * A

        # helpful to store dtype once constructed
        self.dtype = A.dtype
        return A

    def __repr__(self):
        s = "<Lazy({}, shape={}{}{})>"

        s_dtype = (f', dtype={self.dtype}' if self.dtype is not None else '')
        s_factor = (
            f', factor={self.factor}' if self.factor is not None else '')

        return s.format(self.fn.__name__, self.shape, s_dtype, s_factor)



"""
Quantum Information for Many-Body calculations.
"""
import warnings

# some useful math
from math import pi, cos, sin, tan, exp, log, log2, sqrt

# Core functions
from .core import (
    qarray,
    prod,
    isket,
    isbra,
    isop,
    isvec,
    issparse,
    isdense,
    isreal,
    isherm,
    ispos,
    mul,
    dag,
    dot,
    vdot,
    rdot,
    ldmul,
    rdmul,
    outer,
    explt,
    get_thread_pool,
    normalize,
    chop,
    quimbify,
    qu,
    ket,
    bra,
    dop,
    sparse,
    infer_size,
    trace,
    identity,
    eye,
    speye,
    dim_map,
    dim_compress,
    kron,
    kronpow,
    ikron,
    pkron,
    permute,
    itrace,
    partial_trace,
    expectation,
    expec,
    nmlz,
    tr,
    ptr,
)

# Linear algebra functions
from .linalg.base_linalg import (
    eigensystem,
    eig,
    eigh,
    eigvals,
    eigvalsh,
    eigvecs,
    eigvecsh,
    eigensystem_partial,
    groundstate,
    groundenergy,
    bound_spectrum,
    eigh_window,
    eigvalsh_window,
    eigvecsh_window,
    svd,
    svds,
    norm,
    expm,
    sqrtm,
    expm_multiply,
    Lazy,
)
from .linalg.rand_linalg import rsvd, estimate_rank
from .linalg.mpi_launcher import get_mpi_pool

# Generating objects
from .gen.operators import (
    spin_operator,
    pauli,
    hadamard,
    phase_gate,
    S_gate,
    T_gate,
    U_gate,
    rotation,
    Rx,
    Ry,
    Rz,
    Xsqrt,
    Ysqrt,
    Zsqrt,
    Wsqrt,
    swap,
    iswap,
    fsim,
    controlled,
    CNOT,
    cX,
    cY,
    cZ,
    ham_heis,
    ham_ising,
    ham_XY,
    ham_XXZ,
    ham_j1j2,
    ham_mbl,
    ham_heis_2D,
    zspin_projector,
    create,
    destroy,
    num,
    ham_hubbard_hardcore,
)
from .gen.states import (
    basis_vec,
    up,
    zplus,
    down,
    zminus,
    plus,
    xplus,
    minus,
    xminus,
    yplus,
    yminus,
    bloch_state,
    bell_state,
    singlet,
    thermal_state,
    neel_state,
    singlet_pairs,
    werner_state,
    ghz_state,
    w_state,
    levi_civita,
    perm_state,
    graph_state_1d,
    computational_state,
)
from .gen.rand import (
    randn,
    rand,
    rand_matrix,
    rand_herm,
    rand_pos,
    rand_rho,
    rand_ket,
    rand_uni,
    rand_haar_state,
    gen_rand_haar_states,
    rand_mix,
    rand_product_state,
    rand_matrix_product_state,
    rand_mps,
    rand_seperable,
    rand_iso,
    rand_mera,
    seed_rand,
    set_rand_bitgen,
)

# Functions for calculating properties
from .calc import (
    fidelity,
    purify,
    entropy,
    entropy_subsys,
    mutual_information,
    mutinf,
    mutinf_subsys,
    schmidt_gap,
    tr_sqrt,
    tr_sqrt_subsys,
    partial_transpose,
    negativity,
    logarithmic_negativity,
    logneg,
    logneg_subsys,
    concurrence,
    one_way_classical_information,
    quantum_discord,
    trace_distance,
    cprint,
    decomp,
    pauli_decomp,
    bell_decomp,
    correlation,
    pauli_correlations,
    ent_cross_matrix,
    qid,
    is_degenerate,
    is_eigenvector,
    page_entropy,
    heisenberg_energy,
    dephase,
    kraus_op,
    projector,
    measure,
    simulate_counts,
)

# Evolution class and methods
from .evo import Evolution

from .linalg.approx_spectral import (
    approx_spectral_function,
    tr_abs_approx,
    tr_exp_approx,
    tr_sqrt_approx,
    tr_xlogx_approx,
    entropy_subsys_approx,
    logneg_subsys_approx,
    negativity_subsys_approx,
    xlogx,
)
from .utils import (
    save_to_disk,
    load_from_disk,
)


# versioneer
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


warnings.filterwarnings('ignore', message='Caching is not available when ')


__all__ = [
    # Accel ----------------------------------------------------------------- #
    'qarray',
    'prod',
    'isket',
    'isbra',
    'isop',
    'isvec',
    'issparse',
    'isdense',
    'isreal',
    'isherm',
    'ispos',
    'mul',
    'dag',
    'dot',
    'vdot',
    'rdot',
    'ldmul',
    'rdmul',
    'outer',
    'explt',
    # Core ------------------------------------------------------------------ #
    'normalize',
    'chop',
    'quimbify',
    'qu',
    'ket',
    'bra',
    'dop',
    'sparse',
    'infer_size',
    'trace',
    'identity',
    'eye',
    'speye',
    'dim_map',
    'dim_compress',
    'kron',
    'kronpow',
    'ikron',
    'pkron',
    'permute',
    'itrace',
    'partial_trace',
    'expectation',
    'expec',
    'nmlz',
    'tr',
    'ptr',
    # Linalg ---------------------------------------------------------------- #
    'eigensystem',
    'eig',
    'eigh',
    'eigvals',
    'eigvalsh',
    'eigvecs',
    'eigvecsh',
    'eigensystem_partial',
    'groundstate',
    'groundenergy',
    'bound_spectrum',
    'eigh_window',
    'eigvalsh_window',
    'eigvecsh_window',
    'svd',
    'svds',
    'norm',
    'Lazy',
    'rsvd',
    'estimate_rank',
    # Gen ------------------------------------------------------------------- #
    'spin_operator',
    'pauli',
    'hadamard',
    'phase_gate',
    'T_gate',
    'S_gate',
    'U_gate',
    'rotation',
    'Rx',
    'Ry',
    'Rz',
    'Xsqrt',
    'Ysqrt',
    'Zsqrt',
    'Wsqrt',
    'swap',
    'iswap',
    'fsim',
    'controlled',
    'CNOT',
    'cX',
    'cY',
    'cZ',
    'ham_heis',
    'ham_ising',
    'ham_XY',
    'ham_XXZ',
    'ham_j1j2',
    'ham_mbl',
    'ham_heis_2D',
    'create',
    'destroy',
    'num',
    'ham_hubbard_hardcore',
    'zspin_projector',
    'basis_vec',
    'up',
    'zplus',
    'down',
    'zminus',
    'plus',
    'xplus',
    'minus',
    'xminus',
    'yplus',
    'yminus',
    'bloch_state',
    'bell_state',
    'singlet',
    'thermal_state',
    'neel_state',
    'singlet_pairs',
    'werner_state',
    'ghz_state',
    'w_state',
    'levi_civita',
    'perm_state',
    'graph_state_1d',
    'rand_matrix',
    'rand_herm',
    'rand_pos',
    'rand_rho',
    'rand_ket',
    'rand_uni',
    'rand_haar_state',
    'gen_rand_haar_states',
    'rand_mix',
    'rand_mps',
    'randn',
    'rand',
    'rand_product_state',
    'rand_matrix_product_state',
    'rand_seperable',
    'rand_iso',
    'rand_mera',
    'seed_rand',
    'set_rand_bitgen',
    'computational_state',
    # Calc ------------------------------------------------------------------ #
    'expm',
    'sqrtm',
    'expm_multiply',
    'fidelity',
    'purify',
    'entropy',
    'entropy_subsys',
    'mutual_information',
    'mutinf',
    'mutinf_subsys',
    'schmidt_gap',
    'tr_sqrt',
    'tr_sqrt_subsys',
    'partial_transpose',
    'negativity',
    'logarithmic_negativity',
    'logneg',
    'logneg_subsys',
    'concurrence',
    'one_way_classical_information',
    'quantum_discord',
    'trace_distance',
    'cprint',
    'decomp',
    'pauli_decomp',
    'bell_decomp',
    'correlation',
    'pauli_correlations',
    'ent_cross_matrix',
    'qid',
    'is_degenerate',
    'is_eigenvector',
    'page_entropy',
    'heisenberg_energy',
    'dephase',
    'kraus_op',
    'projector',
    'measure',
    'simulate_counts',
    # Evo ------------------------------------------------------------------- #
    'Evolution',
    # Approx spectral ------------------------------------------------------- #
    'approx_spectral_function',
    'tr_abs_approx',
    'tr_exp_approx',
    'tr_sqrt_approx',
    'tr_xlogx_approx',
    'entropy_subsys_approx',
    'logneg_subsys_approx',
    'negativity_subsys_approx',
    # Some misc useful math ------------------------------------------------- #
    'pi',
    'cos',
    'sin',
    'tan',
    'exp',
    'log',
    'log2',
    'sqrt',
    'xlogx',
    # Utils ----------------------------------------------------------------- #
    'save_to_disk',
    'load_from_disk',
    'get_thread_pool',
    'get_mpi_pool',
]



from .tensor_core import (
    get_contract_strategy,
    set_contract_strategy,
    contract_strategy,
    get_contract_backend,
    set_contract_backend,
    contract_backend,
    get_tensor_linop_backend,
    set_tensor_linop_backend,
    tensor_linop_backend,
    tensor_contract,
    tensor_split,
    tensor_direct_product,
    rand_uuid,
    bonds,
    bonds_size,
    connect,
    new_bond,
    Tensor,
    TensorNetwork,
    TNLinearOperator1D,
    PTensor,
)
from .tensor_gen import (
    rand_tensor,
    rand_phased,
    MPS_rand_state,
    MPS_product_state,
    MPS_computational_state,
    MPS_rand_computational_state,
    MPS_neel_state,
    MPS_ghz_state,
    MPS_w_state,
    MPS_zero_state,
    MPS_sampler,
    MPO_identity,
    MPO_identity_like,
    MPO_zeros,
    MPO_zeros_like,
    MPO_rand,
    MPO_rand_herm,
    SpinHam,
    MPO_ham_ising,
    MPO_ham_XY,
    MPO_ham_heis,
    MPO_ham_mbl,
    NNI_ham_ising,
    NNI_ham_XY,
    NNI_ham_heis,
    NNI_ham_mbl,
)
from .tensor_1d import (
    MatrixProductState,
    MatrixProductOperator,
    Dense1D,
    SuperOperator1D,
    align_TN_1D,
    expec_TN_1D,
    gate_TN_1D,
    superop_TN_1D,
)
from .tensor_dmrg import (
    MovingEnvironment,
    DMRG,
    DMRG1,
    DMRG2,
    DMRGX,
)
from .tensor_mera import (
    MERA,
)
from .tensor_tebd import (
    TEBD,
)
from .circuit import (
    Circuit,
    CircuitMPS,
    CircuitDense,
)
from .circuit_gen import (
    circ_ansatz_1D_zigzag,
    circ_ansatz_1D_brickwork,
    circ_ansatz_1D_rand,
)

__all__ = (
    "contract_strategy",
    "get_contract_strategy",
    "set_contract_strategy",
    "contract_backend",
    "get_contract_backend",
    "set_contract_backend",
    "tensor_linop_backend",
    "get_tensor_linop_backend",
    "set_tensor_linop_backend",
    "tensor_contract",
    "tensor_split",
    "tensor_direct_product",
    "rand_uuid",
    "bonds",
    "bonds_size",
    "connect",
    "new_bond",
    "Tensor",
    "TensorNetwork",
    "TNLinearOperator1D",
    "PTensor",
    "rand_tensor",
    "rand_phased",
    "MPS_rand_state",
    "MPS_product_state",
    "MPS_computational_state",
    "MPS_rand_computational_state",
    "MPS_neel_state",
    "MPS_ghz_state",
    "MPS_w_state",
    "MPS_zero_state",
    "MPS_sampler",
    "MPO_identity",
    "MPO_identity_like",
    "MPO_zeros",
    "MPO_zeros_like",
    "MPO_rand",
    "MPO_rand_herm",
    "SpinHam",
    "MPO_ham_ising",
    "MPO_ham_XY",
    "MPO_ham_heis",
    "MPO_ham_mbl",
    "NNI_ham_ising",
    "NNI_ham_XY",
    "NNI_ham_heis",
    "NNI_ham_mbl",
    "MatrixProductState",
    "MatrixProductOperator",
    "Dense1D",
    "SuperOperator1D",
    "align_TN_1D",
    "expec_TN_1D",
    "gate_TN_1D",
    "superop_TN_1D",
    "MovingEnvironment",
    "DMRG",
    "DMRG1",
    "DMRG2",
    "DMRGX",
    "MERA",
    "TEBD",
    "Circuit",
    "CircuitMPS",
    "CircuitDense",
    "circ_ansatz_1D_zigzag",
    "circ_ansatz_1D_brickwork",
    "circ_ansatz_1D_rand",
)



"""Backend agnostic array operations.
"""
import itertools

import numpy
from autoray import do, reshape, transpose, dag, infer_backend

from ..core import njit
from ..linalg.base_linalg import norm_fro_dense


def asarray(array):
    if isinstance(array, numpy.matrix) or not hasattr(array, 'shape'):
        return numpy.asarray(array)
    return array


def ndim(array):
    try:
        return array.ndim
    except AttributeError:
        return len(array.shape)


# ------------- miscelleneous other backend agnostic functions -------------- #

def iscomplex(x):
    if not hasattr(x, 'dtype'):
        return isinstance(x, complex)
    return 'complex' in str(x.dtype)


def norm_fro(x):
    if isinstance(x, numpy.ndarray):
        return norm_fro_dense(x.reshape(-1))
    try:
        return do('linalg.norm', reshape(x, [-1]), 2)
    except AttributeError:
        return do('sum', do('multiply', do('conj', x), x)) ** 0.5


def sensibly_scale(x):
    """Take an array and scale it *very* roughly such that random tensor
    networks consisting of such arrays do not have gigantic norms.
    """
    return x / norm_fro(x)**(1.5 / ndim(x))


def _unitize_qr(x):
    """Perform isometrization using the QR decomposition.
    """
    fat = x.shape[0] < x.shape[1]
    if fat:
        x = transpose(x)

    Q = do('linalg.qr', x)[0]
    if fat:
        Q = transpose(Q)

    return Q


def _unitize_svd(x):
    fat = x.shape[0] < x.shape[1]
    if fat:
        x = transpose(x)

    Q = do('linalg.svd', x)[0]
    if fat:
        Q = transpose(Q)

    return Q


def _unitize_exp(x):
    r"""Perform isometrization using the using anti-symmetric matrix
    exponentiation.

    .. math::

            U_A = \exp{A - A^\dagger}

    If ``x`` is rectangular it is completed with zeros first.
    """
    m, n = x.shape
    d = max(m, n)
    x = do('pad', x, [[0, d - m], [0, d - n]], 'constant', constant_values=0.0)
    expx = do('linalg.expm', x - dag(x))
    return expx[:m, :n]


def _unitize_modified_gram_schmidt(A):
    """Perform isometrization explicitly using the modified Gram Schmidt
    procedure.
    """
    m, n = A.shape

    thin = m > n
    if thin:
        A = do('transpose', A)

    Q = []
    for j in range(0, min(m, n)):

        q = A[j, :]
        for i in range(0, j):
            rij = do('tensordot', do('conj', Q[i]), q, 1)
            q = q - rij * Q[i]

        Q.append(q / do('linalg.norm', q, 2))

    Q = do('stack', Q, axis=0, like=A)

    if thin:
        Q = do('transpose', Q)

    return Q


_UNITIZE_METHODS = {
    'qr': _unitize_qr,
    'svd': _unitize_svd,
    'exp': _unitize_exp,
    'mgs': _unitize_modified_gram_schmidt,
}


def unitize(x, method='qr'):
    """Generate a isometric (or unitary if square) matrix from array ``x``.

    Parameters
    ----------
    x : array
        The matrix to generate the isometry from.
    method : {'qr', 'exp', 'mgs'}, optional
        The method used to generate the isometry. Note ``'qr'`` is the fastest
        and most robust but, for example, some libraries cannot back-propagate
        through it.
    """
    return _UNITIZE_METHODS[method](x)


@njit
def _numba_find_diag_axes(x, atol=1e-12):  # pragma: no cover
    """Numba-compiled array diagonal axis finder.

    Parameters
    ----------
    x : numpy.ndarray
        The array to search for diagonal axes.
    atol : float
        The tolerance with which to compare to zero.

    Returns
    -------
    diag_axes : set[tuple[int]]
        The set of pairs of axes which are diagonal.
    """

    # create the set of pairs of matching size axes
    diag_axes = set()
    for d1 in range(x.ndim - 1):
        for d2 in range(d1 + 1, x.ndim):
            if x.shape[d1] == x.shape[d2]:
                diag_axes.add((d1, d2))

    # enumerate through every array entry, eagerly invalidating axis pairs
    for index, val in numpy.ndenumerate(x):
        for d1, d2 in diag_axes:
            if (index[d1] != index[d2]) and (abs(val) > atol):
                diag_axes.remove((d1, d2))

        # all pairs invalid, nothing left to do
        if len(diag_axes) == 0:
            break

    return diag_axes


def find_diag_axes(x, atol=1e-12):
    """Try and find a pair of axes of ``x`` in which it is diagonal.

    Parameters
    ----------
    x : array-like
        The array to search.
    atol : float, optional
        Tolerance with which to compare to zero.

    Returns
    -------
    tuple[int] or None
        The two axes if found else None.

    Examples
    --------

        >>> x = np.array([[[1, 0], [0, 2]],
        ...               [[3, 0], [0, 4]]])
        >>> find_diag_axes(x)
        (1, 2)

    Which means we can reduce ``x`` without loss of information to:

        >>> np.einsum('abb->ab', x)
        array([[1, 2],
               [3, 4]])

    """
    shape = x.shape
    if len(shape) < 2:
        return None

    backend = infer_backend(x)

    # use numba-accelerated version for numpy arrays
    if backend == 'numpy':
        diag_axes = _numba_find_diag_axes(x, atol=atol)
        if diag_axes:
            # make it determinstic
            return min(diag_axes)
        return None
    indxrs = do('indices', shape, like=backend)

    for i, j in itertools.combinations(range(len(shape)), 2):
        if shape[i] != shape[j]:
            continue
        if do('allclose', x[indxrs[i] != indxrs[j]], 0.0,
              atol=atol, like=backend):
            return (i, j)
    return None


@njit
def _numba_find_antidiag_axes(x, atol=1e-12):  # pragma: no cover
    """Numba-compiled array antidiagonal axis finder.

    Parameters
    ----------
    x : numpy.ndarray
        The array to search for anti-diagonal axes.
    atol : float
        The tolerance with which to compare to zero.

    Returns
    -------
    antidiag_axes : set[tuple[int]]
        The set of pairs of axes which are anti-diagonal.
    """

    # create the set of pairs of matching size axes
    antidiag_axes = set()
    for i in range(x.ndim - 1):
        for j in range(i + 1, x.ndim):
            if x.shape[i] == x.shape[j]:
                antidiag_axes.add((i, j))

    # enumerate through every array entry, eagerly invalidating axis pairs
    for index, val in numpy.ndenumerate(x):
        for i, j in antidiag_axes:
            d = x.shape[i]
            if (index[i] != d - 1 - index[j]) and (abs(val) > atol):
                antidiag_axes.remove((i, j))

        # all pairs invalid, nothing left to do
        if len(antidiag_axes) == 0:
            break

    return antidiag_axes


def find_antidiag_axes(x, atol=1e-12):
    """Try and find a pair of axes of ``x`` in which it is anti-diagonal.

    Parameters
    ----------
    x : array-like
        The array to search.
    atol : float, optional
        Tolerance with which to compare to zero.

    Returns
    -------
    tuple[int] or None
        The two axes if found else None.

    Examples
    --------

        >>> x = np.array([[[0, 1], [0, 2]],
        ...               [[3, 0], [4, 0]]])
        >>> find_antidiag_axes(x)
        (0, 2)

    Which means we can reduce ``x`` without loss of information to:

        >>> np.einsum('aba->ab', x[::-1, :, :])
        array([[3, 4],
               [1, 2]])

    as long as we flip the order of dimensions on other tensors corresponding
    to the the same index.
    """
    shape = x.shape
    if len(shape) < 2:
        return None

    backend = infer_backend(x)

    # use numba-accelerated version for numpy arrays
    if backend == 'numpy':
        antidiag_axes = _numba_find_antidiag_axes(x, atol=atol)
        if antidiag_axes:
            # make it determinstic
            return min(antidiag_axes)
        return None

    indxrs = do('indices', shape, like=backend)

    for i, j in itertools.combinations(range(len(shape)), 2):
        di, dj = shape[i], shape[j]
        if di != dj:
            continue
        if do('allclose', x[indxrs[i] != dj - 1 - indxrs[j]], 0.0,
              atol=atol, like=backend):
            return (i, j)
    return None


@njit
def _numba_find_columns(x, atol=1e-12):  # pragma: no cover
    """Numba-compiled single non-zero column axis finder.

    Parameters
    ----------
    x :  array
        The array to search.
    atol : float, optional
        Absolute tolerance to compare to zero with.

    Returns
    -------
    set[tuple[int]]
        Set of pairs (axis, index) defining lone non-zero columns.
    """

    # possible pairings of axis + index
    column_pairs = set()
    for ax, d in enumerate(x.shape):
        for i in range(d):
            column_pairs.add((ax, i))

    # enumerate over all array entries, invalidating potential column pairs
    for index, val in numpy.ndenumerate(x):
        if abs(val) > atol:
            for ax, i in enumerate(index):
                for pax, pi in column_pairs:
                    if ax == pax and pi != i:
                        column_pairs.remove((pax, pi))

        # all potential pairs invalidated
        if not len(column_pairs):
            break

    return column_pairs


def find_columns(x, atol=1e-12):
    """Try and find columns of axes which are zero apart from a single index.

    Parameters
    ----------
    x : array-like
        The array to search.
    atol : float, optional
        Tolerance with which to compare to zero.

    Returns
    -------
    tuple[int] or None
        If found, the first integer is which axis, and the second is which
        column of that axis, else None.

    Examples
    --------

        >>> x = np.array([[[0, 1], [0, 2]],
        ...               [[0, 3], [0, 4]]])
        >>> find_columns(x)
        (2, 1)

    Which means we can happily slice ``x`` without loss of information to:

        >>> x[:, :, 1]
        array([[1, 2],
               [3, 4]])

    """
    shape = x.shape
    if len(shape) < 1:
        return None

    backend = infer_backend(x)

    # use numba-accelerated version for numpy arrays
    if backend == 'numpy':
        columns_pairs = _numba_find_columns(x, atol)
        if columns_pairs:
            return min(columns_pairs)
        return None

    indxrs = do('indices', shape, like=backend)

    for i in range(len(shape)):
        for j in range(shape[i]):
            if do('allclose', x[indxrs[i] != j], 0.0, atol=atol, like=backend):
                return (i, j)

    return None


class PArray:
    """Simple array-like object that lazily generates the actual array by
    calling a function with a set of parameters.

    Parameters
    ----------
    fn : callable
        The function that generates the tensor data from ``params``.
    params : sequence of numbers
        The initial parameters supplied to the generating function like
        ``fn(params)``.

    See Also
    --------
    PTensor
    """

    def __init__(self, fn, params, shape=None):
        self._fn = fn
        self._params = asarray(params)
        self._shape = shape

    @property
    def fn(self):
        return self._fn

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, x):
        self._params = asarray(x)

    @property
    def data(self):
        self._data = self._fn(self._params)
        return self._data

    @property
    def shape(self):
        if self._shape is None:
            self._shape = self.data.shape
        return self._shape

    @property
    def ndim(self):
        return len(self.shape)



"""Functions for generating random quantum objects and states.
"""
import math
import random
from functools import wraps
from numbers import Integral
from itertools import count, chain
from concurrent.futures import wait

import numpy as np
import numpy.random
import scipy.sparse as sp

from ..core import (qarray, dag, dot, rdmul, complex_array, get_thread_pool,
                    _NUM_THREAD_WORKERS, qu, ptr, kron, nmlz, prod,
                    vectorize, pvectorize)


class _RGenHandler:
    """Private object that handles pool of random number generators for
    parallel number generation - seeding them & changing the underlying bit
    generators.
    """

    def __init__(self, initial_seed=None, initial_bitgen=None):
        self.rgs = []
        self.set_seed(initial_seed)
        self.set_bitgen(initial_bitgen)

    def set_bitgen(self, bitgen):
        """Set the core underlying bit-generator.

        Parameters
        ----------
        bitgen : {None, str}
            Which bit generator to use, either from numpy or `randomgen` -
            https://bashtage.github.io/randomgen/bit_generators/index.html.
        """

        if bitgen is None:
            self.gen_fn = numpy.random.default_rng

        else:
            try:
                bg = getattr(numpy.random, bitgen)
            except AttributeError:
                import randomgen
                bg = getattr(randomgen, bitgen)

            def gen(s):
                return numpy.random.Generator(bg(s))

            self.gen_fn = gen

        # delete any old rgens
        self.rgs = []

    def set_seed(self, seed=None):
        """Set the seed for the bit generators.

        Parameters
        ----------
        seed : {None, int}, optional
            Seed supplied to `numpy.random.SeedSequence`. None will randomly
            the generators (default).
        """
        seq = numpy.random.SeedSequence(seed)

        # compute seeds in batches of 4 for perf
        self.seeds = iter(chain.from_iterable(seq.spawn(4) for _ in count()))

        # delete any old rgens
        self.rgs = []

    def get_rgens(self, num_threads):
        """Get a list of the :class:`numpy.random.Generator` instances, having
        made sure there are at least ``num_threads``.

        Parameters
        ----------
        num_threads : int
            The number of generators to return.

        Returns
        -------
        list[numpy.random.Generator]
        """
        num_gens = len(self.rgs)

        if num_gens < num_threads:
            self.rgs.extend(self.gen_fn(next(self.seeds))
                            for _ in range(num_gens, num_threads))

        return self.rgs[:num_threads]


_RG_HANDLER = _RGenHandler()


def seed_rand(seed):
    """See the random number generators, by instantiating a new set of bit
    generators with a 'seed sequence'.
    """
    global _RG_HANDLER
    return _RG_HANDLER.set_seed(seed)


def set_rand_bitgen(bitgen):
    """Set the core bit generator type to use, from either ``numpy`` or
    ``randomgen``.

    Parameters
    ----------
    bitgen : {'PCG64', 'SFC64', 'MT19937', 'Philox', str}
        Which bit generator to use.
    """
    global _RG_HANDLER
    return _RG_HANDLER.set_bitgen(bitgen)


def _get_rgens(num_threads):
    global _RG_HANDLER
    return _RG_HANDLER.get_rgens(num_threads)


def randn(shape=(), dtype=float, scale=1.0, loc=0.0,
          num_threads=None, seed=None, dist='normal'):
    """Fast multithreaded generation of random normally distributed data
    using ``randomgen``.

    Parameters
    ----------
    shape : tuple[int]
        The shape of the output random array.
    dtype : {'complex128', 'float64', 'complex64' 'float32'}, optional
        The data-type of the output array.
    scale : float, optional
        The width of the distribution (standard deviation if
        ``dist='normal'``).
    loc : float, optional
        The location of the distribution (lower limit if
        ``dist='uniform'``).
    num_threads : int, optional
        How many threads to use. If ``None``, decide automatically.
    dist : {'normal', 'uniform', 'exp'}, optional
        Type of random number to generate.
    """
    if seed is not None:
        seed_rand(seed)

    if isinstance(shape, Integral):
        d = shape
        shape = (shape,)
    else:
        d = prod(shape)

    if num_threads is None:
        # only multi-thread for big ``d``
        if d <= 32768:
            num_threads = 1
        else:
            num_threads = _NUM_THREAD_WORKERS

    rgs = _get_rgens(num_threads)

    gen_method = {
        'uniform': 'random',
        'normal': 'standard_normal',
        'exp': 'standard_exponential',
    }.get(dist, dist)

    # sequential generation
    if num_threads <= 1:

        def create(d, dtype):
            out = np.empty(d, dtype)
            getattr(rgs[0], gen_method)(out=out, dtype=dtype)
            return out

    # threaded generation
    else:
        pool = get_thread_pool()
        S = math.ceil(d / num_threads)

        def _fill(gen, out, dtype, first, last):
            getattr(gen, gen_method)(out=out[first:last], dtype=dtype)

        def create(d, dtype):
            out = np.empty(d, dtype)
            # submit thread work
            fs = [
                pool.submit(_fill, gen, out, dtype, i * S, (i + 1) * S)
                for i, gen in enumerate(rgs)
            ]
            wait(fs)
            return out

    if np.issubdtype(dtype, np.floating):
        out = create(d, dtype)

    elif np.issubdtype(dtype, np.complexfloating):
        # need to sum two real arrays if generating complex numbers
        if np.issubdtype(dtype, np.complex64):
            sub_dtype = np.float32
        else:
            sub_dtype = np.float64

        out = complex_array(create(d, sub_dtype), create(d, sub_dtype))

    else:
        raise ValueError(f"dtype {dtype} not understood.")

    if out.dtype != dtype:
        out = out.astype(dtype)

    if scale != 1.0:
        out *= scale
    if loc != 0.0:
        out += loc

    return out.reshape(shape)


@wraps(randn)
def rand(*args, **kwargs):
    kwargs.setdefault('dist', 'uniform')
    return randn(*args, **kwargs)


def random_seed_fn(fn):
    """Modify ``fn`` to take a ``seed`` argument (so as to seed the random
    generators once-only at beginning of function not every ``randn`` call).
    """

    @wraps(fn)
    def wrapped_fn(*args, seed=None, **kwargs):
        if seed is not None:
            seed_rand(seed)
        return fn(*args, **kwargs)

    return wrapped_fn


def _randint(*args, **kwargs):
    return _get_rgens(1)[0].integers(*args, **kwargs)


def _choice(*args, **kwargs):
    return _get_rgens(1)[0].choice(*args, **kwargs)


choice = random_seed_fn(_choice)


@random_seed_fn
def rand_rademacher(shape, scale=1, dtype=float):
    """
    """
    if np.issubdtype(dtype, np.floating):
        entries = np.array([1.0, -1.0]) * scale
        need2convert = dtype not in (float, np.float_)

    elif np.issubdtype(dtype, np.complexfloating):
        entries = np.array([1.0, -1.0, 1.0j, -1.0j]) * scale
        need2convert = dtype not in (complex, np.complex_)

    else:
        raise TypeError(f"dtype {dtype} not understood - should be float or "
                        "complex.")

    x = _choice(entries, shape)
    if need2convert:
        x = x.astype(dtype)

    return x


def _phase_to_complex_base(x):
    return 1j * math.sin(x) + math.cos(x)


_phase_sigs = ['complex64(float32)', 'complex128(float64)']
_phase_to_complex_seq = vectorize(_phase_sigs)(_phase_to_complex_base)
"""Turn array of phases into unit circle complex numbers - sequential.
"""
_phase_to_complex_par = pvectorize(_phase_sigs)(_phase_to_complex_base)
"""Turn array of phases into unit circle complex numbers - parallel.
"""


def phase_to_complex(x):
    if x.size >= 512:
        return _phase_to_complex_par(x)
    # XXX: this is not as fast as numexpr - investigate?
    return _phase_to_complex_seq(x)


@random_seed_fn
def rand_phase(shape, scale=1, dtype=complex):
    """Generate random complex numbers distributed on the unit sphere.
    """
    if not np.issubdtype(dtype, np.complexfloating):
        raise ValueError(f"dtype must be complex, got '{dtype}'.")

    if np.issubdtype(dtype, np.complex64):
        sub_dtype = np.float32
    else:
        sub_dtype = np.float64

    phi = randn(shape, dtype=sub_dtype, scale=2 * math.pi, dist='uniform')
    z = phase_to_complex(phi)
    if scale != 1:
        z *= scale

    return z


def rand_matrix(d, scaled=True, sparse=False, stype='csr',
                density=None, dtype=complex, seed=None):
    """Generate a random matrix of order `d` with normally distributed
    entries. If `scaled` is `True`, then in the limit of large `d` the
    eigenvalues will be distributed on the unit complex disk.

    Parameters
    ----------
    d : int
        Matrix dimension.
    scaled : bool, optional
        Whether to scale the matrices values such that its spectrum
        approximately lies on the unit disk (for dense matrices).
    sparse : bool, optional
        Whether to produce a sparse matrix.
    stype : {'csr', 'csc', 'coo', ...}, optional
        The type of sparse matrix if ``sparse=True``.
    density : float, optional
        Target density of non-zero elements for the sparse matrix. By default
        aims for about 10 entries per row.
    dtype : {complex, float}, optional
        The data type of the matrix elements.

    Returns
    -------
    mat : qarray or sparse matrix
        Random matrix.
    """
    if np.issubdtype(dtype, np.floating):
        iscomplex = False
    elif np.issubdtype(dtype, np.complexfloating):
        iscomplex = True
    else:
        raise TypeError(f"dtype {dtype} not understood - should be "
                        "float or complex.")

    # handle seed manually since standard python random.seed might be called
    if seed is not None:
        seed_rand(seed)

    if sparse:
        # Aim for 10 non-zero values per row, but betwen 1 and d/2
        density = min(10, d / 2) / d if density is None else density
        density = min(max(d**-2, density, ), 1.0)
        nnz = round(density * d * d)

        if density > 0.1:
            # take special care to avoid duplicates
            if seed is not None:
                random.seed(seed)
            ijs = random.sample(range(0, d**2), k=nnz)
        else:
            ijs = _randint(0, d * d, size=nnz)

        # want to sample nnz unique (d, d) pairs without building list
        i, j = np.divmod(ijs, d)

        data = randn(nnz, dtype=dtype)
        mat = sp.coo_matrix((data, (i, j)), shape=(d, d)).asformat(stype)
    else:
        density = 1.0
        mat = qarray(randn((d, d), dtype=dtype))

    if scaled:
        mat /= ((2 if iscomplex else 1) * d * density)**0.5

    return mat


@random_seed_fn
def rand_herm(d, sparse=False, density=None, dtype=complex):
    """Generate a random hermitian operator of order `d` with normally
    distributed entries. In the limit of large `d` the spectrum will be a
    semi-circular distribution between [-1, 1].

    See Also
    --------
    rand_matrix, rand_pos, rand_rho, rand_uni
    """
    if sparse:
        density = 10 / d if density is None else density
        density = min(max(density, d**-2), 1 - d**-2)
        density /= 2  # to account of herm construction

    herm = rand_matrix(d, scaled=True, sparse=sparse,
                       density=density, dtype=dtype)

    if sparse:
        herm.data /= (2**1.5)
    else:
        herm /= (2**1.5)

    herm += dag(herm)

    return herm


@random_seed_fn
def rand_pos(d, sparse=False, density=None, dtype=complex):
    """Generate a random positive operator of size `d`, with normally
    distributed entries. In the limit of large `d` the spectrum will lie
    between [0, 1].

    See Also
    --------
    rand_matrix, rand_herm, rand_rho, rand_uni
    """
    if sparse:
        density = 10 / d if density is None else density
        density = min(max(density, d**-2), 1 - d**-2)
        density = 0.5 * (density / d)**0.5  # to account for pos construction

    pos = rand_matrix(d, scaled=True, sparse=sparse,
                      density=density, dtype=dtype)

    return dot(pos, dag(pos))


@random_seed_fn
def rand_rho(d, sparse=False, density=None, dtype=complex):
    """Generate a random positive operator of size `d` with normally
    distributed entries and unit trace.

    See Also
    --------
    rand_matrix, rand_herm, rand_pos, rand_uni
    """
    return nmlz(rand_pos(d, sparse=sparse, density=density, dtype=dtype))


@random_seed_fn
def rand_uni(d, dtype=complex):
    """Generate a random unitary operator of size `d`, distributed according to
    the Haar measure.

    See Also
    --------
    rand_matrix, rand_herm, rand_pos, rand_rho
    """
    q, r = np.linalg.qr(rand_matrix(d, dtype=dtype))
    r = np.diagonal(r)
    r = r / np.abs(r)  # read-only so not inplace
    return rdmul(q, r)


@random_seed_fn
def rand_ket(d, sparse=False, stype='csr', density=0.01, dtype=complex):
    """Generates a ket of length `d` with normally distributed entries.
    """
    if sparse:
        ket = sp.random(d, 1, format=stype, density=density)
        ket.data = randn((ket.nnz,), dtype=dtype)
    else:
        ket = qarray(randn((d, 1), dtype=dtype))
    return nmlz(ket)


@random_seed_fn
def rand_haar_state(d, dtype=complex):
    """Generate a random state of dimension `d` according to the Haar
    distribution.
    """
    u = rand_uni(d, dtype=dtype)
    return u[:, [0]]


@random_seed_fn
def gen_rand_haar_states(d, reps, dtype=complex):
    """Generate many random Haar states, recycling a random unitary operator
    by using all of its columns (not a good idea?).
    """
    for rep in range(reps):
        cyc = rep % d
        if cyc == 0:
            u = rand_uni(d, dtype=dtype)
        yield u[:, [cyc]]


@random_seed_fn
def rand_mix(d, tr_d_min=None, tr_d_max=None, mode='rand', dtype=complex):
    """Constructs a random mixed state by tracing out a random ket
    where the composite system varies in size between 2 and d. This produces
    a spread of states including more purity but has no real meaning.
    """
    if tr_d_min is None:
        tr_d_min = 2
    if tr_d_max is None:
        tr_d_max = d

    m = _randint(tr_d_min, tr_d_max)
    if mode == 'rand':
        psi = rand_ket(d * m, dtype=dtype)
    elif mode == 'haar':
        psi = rand_haar_state(d * m, dtype=dtype)

    return ptr(psi, [d, m], 0)


@random_seed_fn
def rand_product_state(n, qtype=None, dtype=complex):
    """Generates a ket of `n` many random pure qubits.
    """
    def gen_rand_pure_qubits(n):
        for _ in range(n):
            u, = rand(1)
            v, = rand(1)
            phi = 2 * np.pi * u
            theta = np.arccos(2 * v - 1)
            yield qu([[np.cos(theta / 2.0)],
                      [np.sin(theta / 2.0) * np.exp(1.0j * phi)]],
                     qtype=qtype, dtype=dtype)
    return kron(*gen_rand_pure_qubits(n))


@random_seed_fn
def rand_matrix_product_state(n, bond_dim, phys_dim=2, dtype=complex,
                              cyclic=False, trans_invar=False):
    """Generate a random matrix product state (in dense form, see
    :func:`~quimb.tensor.MPS_rand_state` for tensor network form).

    Parameters
    ----------
    n : int
        Number of sites.
    bond_dim : int
        Dimension of the bond (virtual) indices.
    phys_dim : int, optional
        Physical dimension of each local site, defaults to 2 (qubits).
    cyclic : bool (optional)
        Whether to impose cyclic boundary conditions on the entanglement
        structure.
    trans_invar : bool (optional)
        Whether to generate a translationally invariant state,
        requires cyclic=True.

    Returns
    -------
    ket : qarray
        The random state, with shape (phys_dim**n, 1)

    """
    from quimb.tensor import MPS_rand_state

    mps = MPS_rand_state(n, bond_dim, phys_dim=phys_dim, dtype=dtype,
                         cyclic=cyclic, trans_invar=trans_invar)
    return mps.to_dense()


rand_mps = rand_matrix_product_state


@random_seed_fn
def rand_seperable(dims, num_mix=10, dtype=complex):
    """Generate a random, mixed, seperable state. E.g rand_seperable([2, 2])
    for a mixed two qubit state with no entanglement.

    Parameters
    ----------
        dims : tuple of int
            The local dimensions across which to be seperable.
        num_mix : int, optional
            How many individual product states to sum together, each with
            random weight.

    Returns
    -------
        qarray
            Mixed seperable state.
    """

    def gen_single_sites():
        for dim in dims:
            yield rand_rho(dim, dtype=dtype)

    weights = rand(num_mix)

    def gen_single_states():
        for w in weights:
            yield w * kron(*gen_single_sites())

    return sum(gen_single_states()) / np.sum(weights)


@random_seed_fn
def rand_iso(n, m, dtype=complex):
    """Generate a random isometry of shape ``(n, m)``.
    """
    data = randn((n, m), dtype=dtype)

    q, _ = np.linalg.qr(data if n >= m else data.T)
    q = q.astype(dtype)

    return q if (n >= m) else q.T


@random_seed_fn
def rand_mera(n, invariant=False, dtype=complex):
    """Generate a random mera state of ``n`` qubits, which must be a power
    of 2. This uses ``quimb.tensor``.
    """
    import quimb.tensor as qt

    if invariant:
        constructor = qt.MERA.rand_invar
    else:
        constructor = qt.MERA.rand

    return constructor(n, dtype=dtype).to_dense()



"""Support for globally optimizing tensor networks using tensorflow in either
interactive (eager) mode or compiled (session) mode. For computationally heavy
optimizations (complex contractions) the two modes offer similar performance,
but for small to medium contractions graph mode is significantly faster.
"""

import time
import functools
from collections import namedtuple

import tqdm
import numpy as np


from .tensor_core import TensorNetwork


LazyComplexTF = namedtuple('LazyComplexTF', ['shape', 'real', 'imag'])
"""We need to lazily represent complex tensorflow arrays so that combining the
variable real and complex parts takes place in the gradient tape always when
executing eagerly.
"""


_TENSORFLOW = None
_TENSORFLOW_EAGER = None


def get_tensorflow(eager=None):
    """Get tensorflow and initialize eager mode if tensorflow has not been
    imported yet. If it has, check whether eager mode has been enabled.
    """
    global _TENSORFLOW
    global _TENSORFLOW_EAGER

    if _TENSORFLOW is None:
        # first check if we are importing tensorflow for the first time
        import sys

        try:
            # leave eager decision to whenever tensorflow was imported
            _TENSORFLOW = sys.modules['tensorflow']
            _TENSORFLOW_EAGER = _TENSORFLOW.executing_eagerly()

        except KeyError:
            import tensorflow as tf

            # default to enabling the more intuitive eager mode
            if eager is None:
                eager = True

            if eager:
                tf.compat.v1.enable_eager_execution()

            _TENSORFLOW = tf
            _TENSORFLOW_EAGER = eager

    return _TENSORFLOW


def executing_eagerly():
    global _TENSORFLOW_EAGER
    return _TENSORFLOW_EAGER


def variable_tn(tn):
    """Convert a tensor network's arrays to tensorflow Variables.
    """
    tf = get_tensorflow()
    tf_tn = tn.copy()
    tf_tn.apply_to_arrays(tf.Variable)
    return tf_tn


def constant_tn(tn):
    """Convert a tensor network's arrays to tensorflow constants.
    """
    tf = get_tensorflow()
    tf_tn = tn.copy()
    tf_tn.apply_to_arrays(tf.convert_to_tensor)
    return tf_tn


def parse_network_to_tf(tn, constant_tags):
    tf = get_tensorflow()
    eager = executing_eagerly()

    tn_tf = tn.copy()
    variables = []
    iscomplex = False

    for t in tn_tf:

        # check if tensor has any of the constant tags
        if t.tags & constant_tags:
            t.modify(data=tf.convert_to_tensor(value=t.data))

        # treat re and im parts as separate variables
        elif issubclass(t.dtype.type, np.complexfloating):
            tf_re, tf_im = tf.Variable(t.data.real), tf.Variable(t.data.imag)
            variables.extend((tf_re, tf_im))

            # if eager need to delay complex combination until gradient tape
            if eager:
                tf_data = LazyComplexTF(t.data.shape, tf_re, tf_im)
            else:
                tf_data = tf.complex(tf_re, tf_im)

            t.modify(data=tf_data)
            iscomplex = True

        else:
            tf_data = tf.Variable(t.data)
            variables.append(tf_data)
            t.modify(data=tf_data)

    return tn_tf, variables, iscomplex


def evaluate_lazy_complex(tn_tf):
    """Can call this in a gradient tape context to merge real and imaginary
    parts so far lazily represented.
    """
    tf = get_tensorflow()

    tn_eval = tn_tf.copy()

    for t in tn_eval:
        if isinstance(t.data, LazyComplexTF):
            t.modify(data=tf.complex(t.data.real, t.data.imag))

    return tn_eval


def init_uninit_vars(sess):
    """ Initialize all other trainable variables, i.e. those which are
    uninitialized. Allows ``optimize`` to be interatively stopped and started.
    """
    tf = get_tensorflow()

    # names of variables
    uninit_vars = sess.run(tf.compat.v1.report_uninitialized_variables())
    vars_list = {v.decode('utf-8') for v in uninit_vars}

    # variables themselves
    uninit_vars_tf = [v for v in tf.compat.v1.global_variables()
                      if v.name.split(':')[0] in vars_list]

    sess.run(tf.compat.v1.variables_initializer(var_list=uninit_vars_tf))


class TNOptimizer:
    """Optimize a tensor network's arrays using tensorflow.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to optimize, every tensor in this will be treated
        as a variable.
    loss_fn : callable
        The target loss function to minimize. Should accept ``tn`` (backed by
        tensorflow tensors) or the output of ``norm_fn``.
    norm_fn : callable, optional
        A function to normalize ``tn`` before being passed to ``loss_fn`` and
        also to call on the final, optimized tensor network. This can be used
        to enforce constraints such as normalization or unitarity.
    loss_constants : dict_like, optional
        Extra constant arguments to supply to ``loss_fn`` and be converted to
        tensorflow constant tensors. Can be individual arrays or tensor
        networks.
    loss_kwargs : dict_like, optional
        Other kwargs to supply to ``loss_fn`` that are not arrays or tensor
        networks.
    constant_tags : sequence of str, optional
        Treat any tensors *within* ``tn`` with these tags as constant.
    optimizer : str, optional
        Which optimizer to use, default: ``'AdamOptimizer'``.
        This should be an optimizer that can be found in the
        ``tensorflow.train`` submodule. The exception is if ``scipy`` is
        supplied the ``tf.contrib.opt.ScipyOptimizerInterface`` will be used,
        the options of which should be specified in the ``optimize`` call.
        This is mainly useful for the default ``'L-BFGS-B'`` algorithm.
    learning_rate : float, optional
        The learning rate to apply to use the optimizer with. You can
        dynamically change this between ``optimize`` calls.
    learning_decay_steps : int, optional
        How many steps to decay the learning rate over.
    learning_decay_rate : float, optional
        How much to decay the learning rate over ``learning_decay_steps``.
    loss_target : float, optional
        If supplied, stop optimizing once this loss is reached.
    progbar : bool, optional
        Whether to show live progress of the optimization.


    Examples
    --------

    .. code:: python3

        import quimb.tensor as qtn
        from quimb.tensor.optimize_tensorflow import TNOptimizer

        # can leave this out, and quimb will invoke tensorflow in
        # eager mode, sacrificing some efficiency
        import tensorflow as tf
        sess = tf.InteractiveSession()

    Variational compression of cyclic MPS, halving its bond dimension:

    .. code:: python3

        # the target state we want to compress
        targ = qtn.MPS_rand_state(n=20, bond_dim=16,
                                  cyclic=True, dtype='complex64')

        # our initial guess for the compressed state
        psi0 = qtn.MPS_rand_state(n=20, bond_dim=8,
                                  cyclic=True, dtype='complex64')

    .. code:: python3

        >>> targ.show()
         16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16 16
        +--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--+
           |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |

        >>> psi0.show()
         8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8 8
        +-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-+
          | | | | | | | | | | | | | | | | | | | |

        >>> abs(psi0.H @ targ)
        0.0012382069

    .. code:: python3

        # the thing (scalar quantity) we want to *minimize*
        def abs_overlap(psi, target):
            overlap = (psi.H & target) ^ all
            return -abs(overlap)

        # also need to define function that keeps us in the normalized manifold
        def psi_normalize(psi):
            nfactor = (psi.H & psi) ^ all
            return psi / nfactor**0.5

    Now we can set-up the optimizer and run it:

    .. code:: python3

        tnopt = TNOptimizer(
            tn=psi0,
            loss_fn=abs_overlap,
            loss_constants={'target': targ},
            norm_fn=psi_normalize,
            optimizer='Adam',
        )

    .. code:: python3

        >>> psif = tnopt.optimize(400)
        -0.7474728226661682: 100%|███████████| 400/400 [00:07<00:00, 51.74it/s]

        >>> abs(psif.H @ targ)
        0.747473

    You could alter the learning rate and keep running this if satisfactory
    results are not achieved. Note we don't here expect this to reach 1.0 as
    for random MPS all the singular values are likely significant and so
    compressing always involves some loss of fidelity.
    """

    def __init__(
        self,
        tn,
        loss_fn,
        norm_fn=None,
        loss_constants=None,
        loss_kwargs=None,
        constant_tags=None,
        optimizer='AdamOptimizer',
        learning_rate=0.1,
        learning_decay_steps=100,
        learning_decay_rate=0.5,
        loss_target=None,
        progbar=True
    ):
        tf = get_tensorflow()

        self.constant_tags = (set() if constant_tags is None
                              else set(constant_tags))

        # make tensorflow version of network and gather variables etc.
        self.tn_opt, self.variables, self.iscomplex = parse_network_to_tf(
            tn, self.constant_tags
        )

        self.progbar = progbar
        self.loss_target = loss_target

        # set dynamic learning rate
        self.global_step = tf.compat.v1.train.get_or_create_global_step()
        self.learning_rate = tf.compat.v1.train.exponential_decay(
            learning_rate=learning_rate, global_step=self.global_step,
            decay_steps=learning_decay_steps, decay_rate=learning_decay_rate,
        )

        # use identity if no nomalization required
        if norm_fn is None:
            def norm_fn(x):
                return x

        self.norm_fn = norm_fn
        self.loss_constants = {}
        if loss_constants is not None:
            for k, v in loss_constants.items():
                # check if tensor network supplied
                if isinstance(v, TensorNetwork):
                    # convert it to constant tensorflow TN
                    self.loss_constants[k] = constant_tn(v)
                else:
                    self.loss_constants[k] = tf.convert_to_tensor(value=v)
        self.loss_kwargs = {} if loss_kwargs is None else dict(loss_kwargs)
        self.loss_fn = functools.partial(
            loss_fn, **self.loss_constants, **self.loss_kwargs
        )
        self.loss = None
        self._n = 0

        if optimizer == 'scipy':
            if executing_eagerly():
                raise ValueError("The tensorflow ``'scipy'`` interface is not "
                                 "available when executing eagerly.")

            self.optimizer = 'scipy'
            return

        elif isinstance(optimizer, str):
            if 'Optimizer' not in optimizer:
                optimizer += 'Optimizer'
            self.optimizer = getattr(
                tf.compat.v1.train, optimizer)(self.learning_rate)
        else:
            self.optimizer = optimizer

        if not executing_eagerly():
            # generate the loss and optimizer computational graphs
            self.loss_op = self.loss_fn(self.norm_fn(self.tn_opt))
            self.train_op = self.optimizer.minimize(
                self.loss_op, global_step=self.global_step)

    @property
    def nevals(self):
        """The number of gradient evaluations.
        """
        return self._n

    def _maybe_start_timer(self, max_time):
        if max_time is None:
            return
        else:
            self._time_start = time.time()

    def _time_should_stop(self, max_time):
        if max_time is None:
            return False
        else:
            return (time.time() - self._time_start) > max_time

    def _get_tn_opt_numpy_eager(self):
        """Get normalized, numpy version of optimized tensor network, in eager
        mode.
        """
        if self.iscomplex:
            tn = evaluate_lazy_complex(self.tn_opt)
        else:
            tn = self.tn_opt

        tn_opt_numpy = self.norm_fn(tn)
        tn_opt_numpy.apply_to_arrays(lambda x: x.numpy())
        return tn_opt_numpy

    def _get_tn_opt_numpy_graph(self, sess):
        """Get normalized, numpy version of optimized tensor network.
        """
        tn_opt_numpy = self.norm_fn(self.tn_opt)

        # evaluate all arrays as once for performance reasons
        np_arrays = sess.run([t.data for t in tn_opt_numpy])
        for t, data in zip(tn_opt_numpy, np_arrays):
            t.modify(data=data)

        return tn_opt_numpy

    def _optimize_scipy(self, max_steps, method='L-BFGS-B',
                        sess=None, max_time=None, **kwargs):
        tf = get_tensorflow()

        loss_op = self.loss_fn(self.norm_fn(self.tn_opt))

        kwargs['maxfun'] = max_steps - 1
        if sess is None:
            sess = tf.compat.v1.get_default_session()
        init_uninit_vars(sess)

        pbar = tqdm.tqdm(total=max_steps, disable=not self.progbar)
        self._maybe_start_timer(max_time)
        timed_out = False

        def step_callback(*_, **__):
            if self._time_should_stop(max_time):
                raise TimeoutError

        def loss_callback(loss_val):
            self._n += 1
            self.loss = loss_val
            pbar.set_description(f"{loss_val}")
            pbar.update()
            return self._time_should_stop(max_time)

        try:
            optimizer = tf.contrib.opt.ScipyOptimizerInterface(
                loss_op, options=kwargs, method=method,
            )
            optimizer.minimize(sess, fetches=[loss_op],
                               loss_callback=loss_callback,
                               step_callback=step_callback)
        except TimeoutError:
            timed_out = True
        finally:
            pbar.close()

            if timed_out:
                raise TimeoutError("The scipy optimizer currently has no way "
                                   "of extracting the current tensor values "
                                   "mid execution.")

            return self._get_tn_opt_numpy_graph(sess)

    def _grad(self):
        tf = get_tensorflow()
        # this is the function that computes the loss and tracks gradient
        with tf.GradientTape() as tape:
            if self.iscomplex:
                tn = evaluate_lazy_complex(self.tn_opt)
            else:
                tn = self.tn_opt
            self.loss = self.loss_fn(self.norm_fn(tn))
        return tape.gradient(self.loss, self.variables)

    def _optimize_eager(self, max_steps, max_time=None):
        # perform the optimization with live progress
        pbar = tqdm.tqdm(total=max_steps, disable=not self.progbar)
        self._maybe_start_timer(max_time)
        try:
            for _ in range(max_steps):
                # compute gradient and display loss value
                grads = self._grad()
                pbar.set_description(f"{self.loss}")

                # performing an optimization step
                self.optimizer.apply_gradients(
                    zip(grads, self.variables),
                    global_step=self.global_step,
                )
                pbar.update()
                self._n += 1

                # check stopping criteria
                if (self.loss_target is not None):
                    if self.loss < self.loss_target:
                        break
                if self._time_should_stop(max_time):
                    break

        finally:
            pbar.close()
            return self._get_tn_opt_numpy_eager()

    def _optimize_graph(self, max_steps, max_time=None, sess=None, ):
        if sess is None:
            tf = get_tensorflow()
            sess = tf.compat.v1.get_default_session()

        init_uninit_vars(sess)

        # perform the optimization with live progress
        pbar = tqdm.tqdm(total=max_steps, disable=not self.progbar)
        self._maybe_start_timer(max_time)
        try:
            for _ in range(max_steps):
                _, self.loss = sess.run([self.train_op, self.loss_op])
                pbar.set_description(f"{self.loss}")
                pbar.update()
                self._n += 1

                # check stopping criteria
                if (self.loss_target is not None):
                    if self.loss < self.loss_target:
                        break
                if self._time_should_stop(max_time):
                    break
        except Exception as e:
            print(e)
        finally:
            pbar.close()
            return self._get_tn_opt_numpy_graph(sess)

    def optimize(self, max_steps, max_time=None, **kwargs):
        """
        Returns
        -------
        tn_opt : TensorNetwork
            The optimized tensor network (with arrays converted back to numpy).
        """
        if self.optimizer == 'scipy':
            kwargs.setdefault('gtol', 1e-12)
            return self._optimize_scipy(max_steps, max_time=max_time, **kwargs)
        elif executing_eagerly():
            return self._optimize_eager(max_steps, max_time=max_time, **kwargs)
        else:
            return self._optimize_graph(max_steps, max_time=max_time, **kwargs)



"""Easy and efficient time evolutions.

Contains an evolution class, Evolution to easily and efficiently manage time
evolution of quantum states according to the Schrodinger equation,
and related functions.
"""

import functools

import numpy as np
from scipy.integrate import complex_ode
from scipy.sparse.linalg import LinearOperator

from .core import (qarray, isop, ldmul, rdmul, explt,
                   dot, issparse, qu, eye, dag, make_immutable)
from .linalg.base_linalg import eigh, norm, expm_multiply
from .linalg.approx_spectral import norm_fro_approx
from .utils import continuous_progbar, progbar


# --------------------------------------------------------------------------- #
# Quantum evolution equations                                                 #
# --------------------------------------------------------------------------- #
#
# This are mostly just to be used internally with the integrators


def schrodinger_eq_ket(ham):
    """Wavefunction schrodinger equation.

    Parameters
    ----------
    ham : operator
        Time-independant Hamiltonian governing evolution.

    Returns
    -------
    psi_dot(t, y) : callable
        Function to calculate psi_dot(t) at psi(t).
    """
    def psi_dot(_, y):
        return -1.0j * dot(ham, y)

    return psi_dot


def schrodinger_eq_ket_timedep(ham):
    """Wavefunction time dependent schrodinger equation.

    Parameters
    ----------
    ham : callable
        Time-dependant Hamiltonian governing evolution, such that ``ham(t)``
        returns an operator representation of the Hamiltonian at time ``t``.

    Returns
    -------
    psi_dot(t, y) : callable
        Function to calculate psi_dot(t) at psi(t).
    """
    def psi_dot(t, y):
        return -1.0j * dot(ham(t), y)

    return psi_dot


def schrodinger_eq_dop(ham):
    """Density operator schrodinger equation, but with flattened input/output.

    Note that this assumes both `ham` and `rho` are hermitian in order to speed
    up the commutator, non-hermitian hamiltonians as used to model loss should
    be treated explicilty or with `schrodinger_eq_dop_vectorized`.

    Parameters
    ----------
    ham : operator
        Time-independant Hamiltonian governing evolution.

    Returns
    -------
    rho_dot(t, y) : callable
        Function to calculate rho_dot(t) at rho(t), input and
        output both in ravelled (1D form).
    """
    d = ham.shape[0]

    def rho_dot(_, y):
        hrho = dot(ham, y.reshape(d, d))
        return -1.0j * (hrho - hrho.T.conj()).reshape(-1)

    return rho_dot


def schrodinger_eq_dop_timedep(ham):
    """Time dependent density operator schrodinger equation, but with flattened
    input/output.

    Note that this assumes both `ham(t)` and `rho` are hermitian in order to
    speed up the commutator, non-hermitian hamiltonians as used to model loss
    should be treated explicilty or with `schrodinger_eq_dop_vectorized`.

    Parameters
    ----------
    ham : callable
        Time-dependant Hamiltonian governing evolution, such that ``ham(t)``
        returns an operator representation of the Hamiltonian at time ``t``.

    Returns
    -------
    rho_dot(t, y) : callable
        Function to calculate rho_dot(t) at rho(t), input and
        output both in ravelled (1D form).
    """
    d = ham(0).shape[0]

    def rho_dot(t, y):
        hrho = dot(ham(t), y.reshape(d, d))
        return -1.0j * (hrho - hrho.T.conj()).reshape(-1)

    return rho_dot


def schrodinger_eq_dop_vectorized(ham):
    """Density operator schrodinger equation, but with flattened input/output
    and vectorised superoperator mode (no reshaping required).

    Note that this is probably only more efficient for sparse Hamiltonians.

    Parameters
    ----------
    ham: time-independant hamiltonian governing evolution

    Returns
    -------
    rho_dot(t, y) : callable
        Function to calculate rho_dot(t) at rho(t), input and
        output both in ravelled (1D form).
    """
    d = ham.shape[0]
    sparse = issparse(ham)
    idt = eye(d, sparse=sparse)
    evo_superop = -1.0j * ((ham & idt) - (idt & ham.T))

    def rho_dot(_, y):
        return dot(evo_superop, y)
    return rho_dot


def lindblad_eq(ham, ls, gamma):
    """Lindblad equation, but with flattened input/output.

    Parameters
    ----------
    ham : operator
        Time-independant hamiltonian governing evolution.
    ls : sequence of matrices
        Lindblad operators.
    gamma : float
        Dampening strength.

    Returns
    -------
    rho_dot(t, y) : callable
        Function to calculate rho_dot(t) at rho(t), input and
        output both in ravelled (1D form).
    """
    d = ham.shape[0]
    lls = tuple(dot(dag(l), l) for l in ls)

    def gen_l_terms(rho):
        for l, ll in zip(ls, lls):
            yield (dot(l, dot(rho, dag(l))) -
                   0.5 * (dot(rho, ll) + dot(ll, rho)))

    def rho_dot(_, y):
        rho = y.reshape(d, d)
        rho_d = dot(ham, rho)
        rho_d -= rho_d.T.conj()
        rho_d *= -1.0j
        rho_d += gamma * sum(gen_l_terms(rho))
        return np.asarray(rho_d).reshape(-1)

    return rho_dot


def lindblad_eq_vectorized(ham, ls, gamma, sparse=False):
    """Lindblad equation, but with flattened input/output and vectorised
    superoperation mode (no reshaping required).

    Parameters
    ----------
    ham : operator
        Time-independant hamiltonian governing evolution.
    ls : sequence of matrices
        Lindblad operators.
    gamma : float
        Dampening strength.

    Returns
    -------
    rho_dot(t, y) : callable
        Function to calculate rho_dot(t) at rho(t), input and
        output both in ravelled (1D form).
    """
    d = ham.shape[0]
    ham_sparse = issparse(ham) or sparse
    idt = eye(d, sparse=ham_sparse)
    evo_superop = -1.0j * ((ham & idt) - (idt & ham.T))

    def gen_lb_terms():
        for l in ls:
            lb_sparse = issparse(l) or sparse
            idt = eye(d, sparse=lb_sparse)
            yield ((l & l.conj()) - 0.5 * ((idt & dot(dag(l), l).T) +
                                           (dot(dag(l), l) & idt)))

    evo_superop += gamma * sum(gen_lb_terms())

    def rho_dot(_, y):
        return dot(evo_superop, y)
    return rho_dot


def _calc_evo_eq(isdop, issparse, isopen=False, timedep=False):
    """Choose an appropirate dynamical equation to evolve with.
    """
    eq_chooser = {
        (0, 0, 0, 0): schrodinger_eq_ket,
        (0, 1, 0, 0): schrodinger_eq_ket,
        (1, 0, 0, 0): schrodinger_eq_dop,
        (1, 1, 0, 0): schrodinger_eq_dop_vectorized,
        (1, 0, 1, 0): lindblad_eq,
        (1, 1, 1, 0): lindblad_eq_vectorized,
        # time-dependent
        (0, 0, 0, 1): schrodinger_eq_ket_timedep,
        (0, 1, 0, 1): schrodinger_eq_ket_timedep,
        (1, 0, 0, 1): schrodinger_eq_dop_timedep,
        (1, 1, 0, 1): schrodinger_eq_dop_timedep,
    }
    return eq_chooser[(isdop, issparse, isopen, timedep)]


# --------------------------------------------------------------------------- #
# Quantum Evolution Class                                                     #
# --------------------------------------------------------------------------- #

class Evolution(object):
    """A class for evolving quantum systems according to Schrodinger equation.

    The evolution can be performed in a number of ways:

        - diagonalise the Hamiltonian (or use already diagonalised system).
        - integrate the complex ODE, that is, the Schrodinger equation, using
          scipy. Here either a mid- or high-order Dormand-Prince adaptive
          time stepping scheme is used (see
          :class:`scipy.integrate.complex_ode`).

    Parameters
    ----------
    p0 : quantum state
        Inital state, either vector or operator. If vector, converted to ket.
    ham : operator, tuple (1d array, operator), or callable
        Governing Hamiltonian, if tuple then assumed to contain
        ``(eigvals, eigvecs)`` of presolved system. If callable (but not a
        SciPy ``LinearOperator``), assume a time-dependent hamiltonian such
        that ``ham(t)`` is the Hamiltonian at time ``t``. In this case, the
        latest call to ``ham`` will be cached (and made immutable) in case it
        is needed by callbacks passed to ``compute``.
    t0 : float, optional
        Initial time (i.e. time of state ``p0``), defaults to zero.
    compute : callable, or dict of callable, optional
        Function(s) to compute on the state at each time step. Function(s)
        should take args (t, pt) or (t, pt, ham) if the Hamiltonian is
        required. If ham is required, it will be passed in to the function
        exactly as given to this ``Evolution`` instance, except if ``method``
        is ``'solve'``, in which case it will be passed in as the solved system
        ``(eigvals, eigvecs)``. If supplied with:

            - single callable : ``Evolution.results`` will contain the results
              as a list,
            - dict of callables : ``Evolution.results`` will contain the
              results as a dict of lists with corresponding keys to those
              given in ``compute``.

    method : {'integrate', 'solve', 'expm'}
        How to evolve the system:

            - ``'integrate'``: use definite integration. Get system at each
              time step, only need action of Hamiltonian on state. Generally
              efficient.
            - ``'solve'``: diagonalise dense hamiltonian. Best for small
              systems and allows arbitrary time steps without loss of
              precision.
            - ``'expm'``: compute the evolved state using the action of the
              operator exponential in a 'single shot' style. Only needs action
              of Hamiltonian, for very large systems can use distributed MPI.

    int_small_step : bool, optional
        If ``method='integrate'``, whether to use a low or high order
        integrator to give naturally small or large steps.
    expm_backend : {'auto', 'scipy', 'slepc'}
        How to perform the expm_multiply function if ``method='expm'``. Can
        further specifiy ``'slepc-krylov'``, or ``'slepc-expokit'``.
    expm_opts : dict
        Supplied to :func:`~quimb.linalg.base_linalg.expm_multiply`
        function if ``method='expm'``.
    progbar : bool, optional
        Whether to show a progress bar when calling ``at_times`` or integrating
        with the ``update_to`` method.
    """

    def __init__(self, p0, ham, t0=0,
                 compute=None,
                 method='integrate',
                 int_small_step=False,
                 expm_backend='AUTO',
                 expm_opts=None,
                 progbar=False):

        self._p0 = qu(p0)
        self._t = self.t0 = t0
        self._isdop = isop(self._p0)  # Density operator evolution?
        self._d = p0.shape[0]  # Hilbert space dimension
        self._progbar = progbar

        self._timedep = callable(ham) and not isinstance(ham, LinearOperator)

        if self._timedep:
            # cache the time-dependent Hamiltonian in case callbacks use it
            noncacheing_ham = ham

            @functools.lru_cache(1)
            def ham(t):
                Ht = noncacheing_ham(t)
                if not isinstance(Ht, LinearOperator):
                    make_immutable(Ht)
                return Ht

        self._setup_callback(compute)
        self._method = method

        if method == 'solve' or isinstance(ham, (tuple, list)):
            if isinstance(ham, LinearOperator):
                raise TypeError("You can't use the 'solve' method "
                                "with an abstract linear operator "
                                "Hamiltonian.")
            elif self._timedep:
                raise TypeError("You can't use the 'solve' method "
                                "with a time-dependent Hamiltonian.")
            self._ham = ham
            self._setup_solved_ham()

        elif method == 'integrate':
            self._start_integrator(ham, int_small_step)
            self._ham = ham
        elif method == 'expm':
            if isinstance(ham, LinearOperator):
                raise TypeError("You can't use the 'expm' method "
                                "with an abstract linear operator "
                                "Hamiltonian.")
            elif self._timedep:
                raise TypeError("You can't use the 'expm' method "
                                "with a time-dependent Hamiltonian.")
            self._update_method = self._update_to_expm_ket
            self._pt = self._p0
            self.expm_backend = expm_backend
            self.expm_opts = {} if expm_opts is None else dict(expm_opts)
            self._ham = ham
        else:
            raise ValueError(f"Did not understand evolution "
                             "method: '{method}'.")

    def _setup_callback(self, fn):
        """Setup callbacks in the correct place to compute into _results
        """
        # if fn is None there is no callback, but possible make a dummy
        #   one if progbar needs updating
        if fn is None:
            step_callback = None
            if not self._progbar:
                int_step_callback = None
            else:
                def int_step_callback(t, y, H):
                    pass

        # else fn is a dict of callbacks or a single callback
        else:
            # try just passing time and state to a single callback function
            # if it fails, pass time, state and hamiltonian
            def single_callback(fn, t, pt, H):
                try:
                    fn_result = fn(t, pt)
                except TypeError as e:
                    if 'positional' in e.args[0]:
                        fn_result = fn(t, pt, H)
                    else:
                        raise
                return fn_result

            # dict of funcs input -> dict of funcs output
            if isinstance(fn, dict):
                self._results = {k: [] for k in fn}

                def step_callback(t, pt, H):
                    for k, v in fn.items():
                        fn_result = single_callback(v, t, pt, H)
                        self._results[k].append(fn_result)

            # else results -> single list of outputs of fn
            else:
                self._results = []

                def step_callback(t, pt, H):
                    fn_result = single_callback(fn, t, pt, H)
                    self._results.append(fn_result)

            # For the integration callback, additionally need to convert
            #   back to 'quantum' (column vector) form

            def int_step_callback(t, y, H):
                pt = qarray(y.reshape(self._d, -1))
                step_callback(t, pt, H)

        self._step_callback = step_callback
        self._int_step_callback = int_step_callback

    def _setup_solved_ham(self):
        """Solve the hamiltonian if needed and find the initial state
        in the energy eigenbasis for quick evolution later.
        """
        # See if already solved from tuple
        try:
            evals, evecs = self._ham
            self._method = 'solve'
        except ValueError:
            evals, evecs = eigh(self._ham.A)
            self._ham = (evals, evecs)

        # Find initial state in energy eigenbasis at t0
        if self._isdop:
            self.pe0 = dot(dag(evecs), dot(self._p0, evecs))
            self._update_method = self._update_to_solved_dop
        else:
            self.pe0 = dot(dag(evecs), self._p0)
            self._update_method = self._update_to_solved_ket

        # Current state (start with same as initial)
        self._pt = self._p0

    def _start_integrator(self, ham, small_step):
        """Initialize a stepping integrator.
        """
        if self._timedep:
            H0 = ham(0.0)
        else:
            H0 = ham

        # set complex ode with governing equation
        evo_eq = _calc_evo_eq(self._isdop, issparse(H0), False, self._timedep)

        self._stepper = complex_ode(evo_eq(ham))

        # 5th order stpper or 8th order stepper
        int_mthd, step_fct = ('dopri5', 150) if small_step else ('dop853', 50)
        if isinstance(H0, LinearOperator):
            # approx norm doesn't need to be very accurate
            nrm0 = norm_fro_approx(H0, tol=0.1)
        else:
            nrm0 = norm(H0, 'f')
        first_step = nrm0 / step_fct

        self._stepper.set_integrator(int_mthd, nsteps=0, first_step=first_step)

        # Set step_callback to be evaluated with args (t, y) at each step
        if self._int_step_callback is not None:
            def solout(t, y):
                self._int_step_callback(t, y, self._ham)
            self._stepper.set_solout(solout)

        self._stepper.set_initial_value(self._p0.A.reshape(-1), self.t0)

        # assign the correct update_to method
        self._update_method = self._update_to_integrate

    # Methods for updating the simulation ----------------------------------- #

    def _update_to_expm_ket(self, t):
        """Update the simulation to time ``t``, without explicitly computing
        the operator exponential itself.
        """
        factor = -1j * (t - self.t)
        self._pt = expm_multiply(factor * self._ham, self._pt,
                                 backend=self.expm_backend, **self.expm_opts)
        self._t = t

        # compute any callbacks into -> self._results
        if self._step_callback is not None:
            self._step_callback(t, self._pt, self._ham)

    def _update_to_solved_ket(self, t):
        """Update simulation consisting of a solved hamiltonian and a
        wavefunction to time `t`.
        """
        self._t = t
        evals, evecs = self._ham
        lt = explt(evals, t - self.t0)
        self._pt = evecs @ ldmul(lt, self.pe0)

        # compute any callbacks into -> self._results
        if self._step_callback is not None:
            self._step_callback(t, self._pt, self._ham)

    def _update_to_solved_dop(self, t):
        """Update simulation consisting of a solved hamiltonian and a
        density operator to time `t`.
        """
        self._t = t
        evals, evecs = self._ham
        lt = explt(evals, t - self.t0)
        lvpvl = rdmul(ldmul(lt, self.pe0), lt.conj())
        self._pt = evecs @ (lvpvl @ dag(evecs))

        # compute any callbacks into -> self._results
        if self._step_callback is not None:
            self._step_callback(t, self._pt, self._ham)

    def _update_to_integrate(self, t):
        """Update simulation consisting of unsolved hamiltonian.
        """
        self._stepper.integrate(t)

    def update_to(self, t):
        """Update the simulation to time ``t`` using relevant method.

        Parameters
        ----------
        t : float
            Time to update the evolution to.
        """
        if self._progbar and hasattr(self, '_stepper'):

            with continuous_progbar(self.t, t) as pbar:
                if self._int_step_callback is not None:
                    def solout(t, y):
                        self._int_step_callback(t, y, self._ham)
                        pbar.cupdate(t)
                else:
                    def solout(t, _):
                        pbar.cupdate(t)

                self._stepper.set_solout(solout)
                self._update_method(t)
        else:
            self._update_method(t)

    def at_times(self, ts):
        """Generator expression to yield state af list of times.

        Parameters
        ----------
        ts : sequence of floats
            Times at which to evolve to, then yield the state.

        Yields
        ------
        pt : quantum state
            Quantum state of evolution at next time in ``ts``.

        Notes
        -----
        If integrating, currently any compute callbacks will be called at every
        *integration* step, not just the times `ts` -- i.e. in general
        len(Evolution.results) != len(ts) and if the adaptive step times are
        needed they should be added as a callback, e.g.
        ``compute['t'] = lambda t, _: return t``.
        """
        if self._progbar:
            ts = progbar(ts)

        for t in ts:
            self._update_method(t)
            yield self.pt

    # Simulation properties ------------------------------------------------- #

    @property
    def t(self):
        """float : Current time of simulation.
        """
        return self._stepper.t if self._method == 'integrate' else self._t

    @property
    def pt(self):
        """quantum state : State of the system at the current time (t).
        """
        if self._method == 'integrate':
            return qarray(self._stepper.y.reshape(self._d, -1))
        else:
            return self._pt

    @property
    def results(self):
        """list, or dict of lists, optional : Results of the compute
        callback(s) for each time step.
        """
        return self._results



"""Functions for generating quantum operators.
"""
from operator import add
import math
import functools
import itertools
import operator

from cytoolz import isiterable, concat, unique
import numpy as np
import scipy.sparse as sp
from scipy.special import comb

from ..core import (qarray, make_immutable, get_thread_pool,
                    par_reduce, isreal, qu, eye, kron, ikron)


# --------------------------------------------------------------------------- #
#                      gates and other simple operators                       #
# --------------------------------------------------------------------------- #

@functools.lru_cache(maxsize=16)
def spin_operator(label, S=1 / 2, **kwargs):
    """Generate a general spin-operator.

    Parameters
    ----------
    label : str
        The type of operator, can be one of six options:
            - ``{'x', 'X'}``, x-spin operator.
            - ``{'y', 'Y'}``, y-spin operator.
            - ``{'z', 'Z'}``, z-spin operator.
            - ``{'+', 'p'}``, Raising operator.
            - ``{'-', 'm'}``, Lowering operator.
            - ``{'i', 'I'}``, identity operator.
    S : float, optional
        The spin of particle to act on, default to spin-1/2.
    kwargs
        Passed to :func:`quimbify`.

    Returns
    -------
    S : immutable operator
        The spin operator.

    See Also
    --------
    pauli

    Examples
    --------
    >>> spin_operator('x')
    qarray([[0. +0.j, 0.5+0.j],
            [0.5+0.j, 0. +0.j]])

    >>> qu.spin_operator('+', S=1)
    qarray([[0.        +0.j, 1.41421356+0.j, 0.        +0.j],
            [0.        +0.j, 0.        +0.j, 1.41421356+0.j],
            [0.        +0.j, 0.        +0.j, 0.        +0.j]])

    >>> qu.spin_operator('Y', sparse=True)
    <2x2 sparse matrix of type '<class 'numpy.complex128'>'
        with 2 stored elements in Compressed Sparse Row format>

    """

    D = int(2 * S + 1)

    op = np.zeros((D, D), dtype=complex)
    ms = np.linspace(S, -S, D)

    label = label.lower()

    if label in {'x', 'y'}:
        for i in range(D - 1):
            c = 0.5 * (S * (S + 1) - (ms[i] * ms[i + 1]))**0.5
            op[i, i + 1] = -1.0j * c if (label == 'y') else c
            op[i + 1, i] = 1.0j * c if (label == 'y') else c

    elif label == 'z':
        for i in range(D):
            op[i, i] = ms[i]

    elif label in {'+', 'p', '-', 'm'}:
        for i in range(D - 1):
            c = (S * (S + 1) - (ms[i] * ms[i + 1]))**0.5
            if label in {'+', 'p'}:
                op[i, i + 1] = c
            else:
                op[i + 1, i] = c

    elif label in {'i', 'I'}:
        np.fill_diagonal(op, 1.0)

    else:
        raise ValueError(f"Label '{label}'' not understood, should be one of "
                         "``['X', 'Y', 'Z', '+', '-', 'I']``.")

    op = qu(np.real_if_close(op), **kwargs)

    make_immutable(op)
    return op


@functools.lru_cache(maxsize=8)
def pauli(xyz, dim=2, **kwargs):
    """Generates the pauli operators for dimension 2 or 3.

    Parameters
    ----------
    xyz : str
        Which spatial direction, upper or lower case from ``{'I', 'X', 'Y',
        'Z'}``.
    dim : int, optional
        Dimension of spin operator (e.g. 3 for spin-1), defaults to 2 for
        spin half.
    kwargs
        Passed to ``quimbify``.

    Returns
    -------
    P : immutable operator
        The pauli operator.

    See Also
    --------
    spin_operator
    """
    xyzmap = {0: 'i', 'i': 'i', 'I': 'i',
              1: 'x', 'x': 'x', 'X': 'x',
              2: 'y', 'y': 'y', 'Y': 'y',
              3: 'z', 'z': 'z', 'Z': 'z'}
    opmap = {('i', 2): lambda: eye(2, **kwargs),
             ('x', 2): lambda: qu([[0, 1],
                                   [1, 0]], **kwargs),
             ('y', 2): lambda: qu([[0, -1j],
                                   [1j, 0]], **kwargs),
             ('z', 2): lambda: qu([[1, 0],
                                   [0, -1]], **kwargs),
             ('i', 3): lambda: eye(3, **kwargs),
             ('x', 3): lambda: qu([[0, 1, 0],
                                   [1, 0, 1],
                                   [0, 1, 0]], **kwargs) / 2**.5,
             ('y', 3): lambda: qu([[0, -1j, 0],
                                   [1j, 0, -1j],
                                   [0, 1j, 0]], **kwargs) / 2**.5,
             ('z', 3): lambda: qu([[1, 0, 0],
                                   [0, 0, 0],
                                   [0, 0, -1]], **kwargs)}
    op = opmap[(xyzmap[xyz], dim)]()

    # Operator is cached, so make sure it cannot be modified
    make_immutable(op)

    return op


@functools.lru_cache(8)
def hadamard(dtype=complex, sparse=False):
    """The Hadamard gate.
    """
    H = qu([[1., 1.],
            [1., -1.]], dtype=dtype, sparse=sparse) / 2**0.5
    make_immutable(H)
    return H


@functools.lru_cache(128)
def phase_gate(phi, dtype=complex, sparse=False):
    """The generalized qubit phase-gate, which adds phase ``phi`` to the
    ``|1>`` state.
    """
    Rp = qu([[1., 0.],
             [0., np.exp(1.0j * phi)]], dtype=dtype, sparse=sparse)
    make_immutable(Rp)
    return Rp


@functools.lru_cache(8)
def T_gate(dtype=complex, sparse=False):
    """The T-gate (pi/8 gate).
    """
    return phase_gate(math.pi / 4, dtype=dtype, sparse=sparse)


@functools.lru_cache(8)
def S_gate(dtype=complex, sparse=False):
    """The S-gate (phase gate).
    """
    return phase_gate(math.pi / 2, dtype=dtype, sparse=sparse)


@functools.lru_cache(128)
def rotation(phi, xyz='Z', dtype=complex, sparse=False):
    """The single qubit rotation gate.
    """
    R = math.cos(phi / 2) * pauli('I') - 1.0j * math.sin(phi / 2) * pauli(xyz)
    R = qu(R, dtype=dtype, sparse=sparse)
    make_immutable(R)
    return R


Rx = functools.partial(rotation, xyz='x')
Ry = functools.partial(rotation, xyz='y')
Rz = functools.partial(rotation, xyz='z')


@functools.lru_cache(128)
def U_gate(theta, phi, lamda, dtype=complex, sparse=False):
    r"""Arbitrary unitary single qubit gate.

    .. math::

        U_3(\theta, \phi, \lambda) =
        \begin{bmatrix}
        \cos(\theta / 2) & - e^{i \lambda} \sin(\theta / 2) \\
        e^{i \phi} \sin(\theta / 2) & e^{i(\lambda + \phi)}\cos(\theta / 2)
        \end{bmatrix}

    Parameters
    ----------
    theta : float
        Angle between 0 and pi.
    phi : float
        Angle between 0 and 2 pi.
    lamba : float
        Angle between 0 and 2 pi.

    Returns
    -------
    U : (2, 2) array
        The unitary matrix, cached.
    """
    from cmath import cos, sin, exp

    c2, s2 = cos(theta / 2), sin(theta / 2)
    U = qu(
        [[c2, -exp(1j * lamda) * s2],
         [exp(1j * phi) * s2, exp(1j * (lamda + phi)) * c2]],
        dtype=dtype, sparse=sparse
    )
    make_immutable(U)
    return U


@functools.lru_cache(4)
def Xsqrt(**qu_opts):
    r"""Rx(pi / 2).

    .. math::

        X^{\frac{1}{2}} =
        \frac{1}{\sqrt{2}}
        \begin{bmatrix}
        1  & - i \\
        - i & 1
        \end{bmatrix}
    """
    return Rx(math.pi / 2, **qu_opts)


@functools.lru_cache(4)
def Ysqrt(**qu_opts):
    r"""Ry(pi / 2).

    .. math::

        Y^{\frac{1}{2}} =
        \frac{1}{\sqrt{2}}
        \begin{bmatrix}
        1 & - 1 \\
        1 & 1
        \end{bmatrix}
    """
    return Ry(math.pi / 2, **qu_opts)


@functools.lru_cache(4)
def Zsqrt(**qu_opts):
    r"""Rz(pi / 2).

    .. math::

        Z^{\frac{1}{2}} =
        \frac{1}{\sqrt{2}}
        \begin{bmatrix}
        1 - i & 0 \\
        0 & 1 + i
        \end{bmatrix}
    """
    return Ry(math.pi / 2, **qu_opts)


@functools.lru_cache(4)
def Wsqrt(**qu_opts):
    r"""R[X + Y](pi / 2).

    .. math::

        W^{\frac{1}{2}} =
        \frac{1}{\sqrt{2}}
        \begin{bmatrix}
        1 & -\sqrt{i} \\
        \sqrt{-i} & 1
        \end{bmatrix}
    """
    return U_gate(math.pi / 2, -math.pi / 4, math.pi / 4, **qu_opts)


@functools.lru_cache(maxsize=8)
def swap(dim=2, dtype=complex, **kwargs):
    """The SWAP operator acting on subsystems of dimension `dim`.
    """
    S = np.identity(dim**2, dtype=dtype)
    S = (S.reshape([dim, dim, dim, dim])
          .transpose([0, 3, 1, 2])
          .reshape([dim**2, dim**2]))
    S = qu(S, dtype=dtype, **kwargs)
    make_immutable(S)
    return S


@functools.lru_cache(maxsize=128)
def fsim(theta, phi, dtype=complex, **kwargs):
    r"""The 'fermionic simulation' gate:

    .. math::

        \mathrm{fsim}(\theta, \phi) =
        \begin{bmatrix}
        1 & 0 & 0 & 0\\
        0 & \cos(\theta) & -i sin(\theta) & 0\\
        0 & -i sin(\theta) & \cos(\theta) & 0\\
        0 & 0 & 0 & \exp(-i \phi)
        \end{bmatrix}

    Note that ``theta`` and ``phi`` should be specified in radians and the sign
    convention with this gate varies. Here for example,
    ``fsim(- pi / 2, 0) == iswap()``.
    """
    from cmath import cos, sin, exp

    a = cos(theta)
    b = -1j * sin(theta)
    c = exp(-1j * phi)

    gate = [[1, 0, 0, 0],
            [0, a, b, 0],
            [0, b, a, 0],
            [0, 0, 0, c]]

    gate = qu(gate, dtype=dtype, **kwargs)
    make_immutable(gate)
    return gate


@functools.lru_cache(maxsize=4)
def iswap(dtype=complex, **kwargs):
    iswap = qu([[1., 0., 0., 0.],
                [0., 0., 1j, 0.],
                [0., 1j, 0., 0.],
                [0., 0., 0., 1.]], dtype=dtype, **kwargs)
    make_immutable(iswap)
    return iswap


@functools.lru_cache(maxsize=16)
def controlled(s, dtype=complex, sparse=False):
    """Construct a controlled pauli gate for two qubits.

    Parameters
    ----------
    s : str
        Which pauli to use, including 'not' aliased to 'x'.
    sparse : bool, optional
        Whether to construct a sparse operator.

    Returns
    -------
    C : immutable operator
        The controlled two-qubit gate operator.
    """
    # alias not and NOT to x
    s = {'NOT': 'x', 'not': 'x'}.get(s, s)
    kws = {'dtype': dtype, 'sparse': sparse}
    op = ((qu([1, 0], qtype='dop', **kws) & eye(2, **kws)) +
          (qu([0, 1], qtype='dop', **kws) & pauli(s, **kws)))
    make_immutable(op)
    return op


@functools.lru_cache(8)
def CNOT(dtype=complex, sparse=False):
    """The controlled-not gate.
    """
    return controlled('not', dtype=dtype, sparse=sparse)


@functools.lru_cache(8)
def cX(dtype=complex, sparse=False):
    """The controlled-X gate.
    """
    return controlled('not', dtype=dtype, sparse=sparse)


@functools.lru_cache(8)
def cY(dtype=complex, sparse=False):
    """The controlled-Y gate.
    """
    return controlled('Y', dtype=dtype, sparse=sparse)


@functools.lru_cache(8)
def cZ(dtype=complex, sparse=False):
    """The controlled-Z gate.
    """
    return controlled('Z', dtype=dtype, sparse=sparse)


# --------------------------------------------------------------------------- #
#                                Hamiltonians                                 #
# --------------------------------------------------------------------------- #


def hamiltonian_builder(fn):
    """Wrap a function to perform some generic postprocessing and take the
    kwargs ``stype`` and ``sparse``. This assumes the core function always
    builds the hamiltonian in sparse form. The wrapper then:

    1. Checks if the operator is real and, if so, discards imaginary part if no
       explicity `dtype` was given
    2. Converts the operator to dense or the correct sparse form
    3. Makes the operator immutable so it can be safely cached
    """

    @functools.wraps(fn)
    def ham_fn(*args, stype='csr', sparse=False, **kwargs):
        H = fn(*args, **kwargs)

        if kwargs.get('dtype', None) is None and isreal(H):
            H = H.real

        if not sparse:
            H = qarray(H.A)
        elif H.format != stype:
            H = H.asformat(stype)

        make_immutable(H)

        return H

    return ham_fn


@functools.lru_cache(maxsize=8)
@hamiltonian_builder
def ham_heis(n, j=1.0, b=0.0, cyclic=True,
             parallel=False, nthreads=None, ownership=None):
    """Constructs the nearest neighbour 1d heisenberg spin-1/2 hamiltonian.

    Parameters
    ----------
    n : int
        Number of spins.
    j : float or tuple(float, float, float), optional
        Coupling constant(s), with convention that positive =
        antiferromagnetic. Can supply scalar for isotropic coupling or
        vector ``(jx, jy, jz)``.
    b : float or tuple(float, float, float), optional
        Magnetic field, defaults to z-direction only if tuple not given.
    cyclic : bool, optional
        Whether to couple the first and last spins.
    sparse : bool, optional
        Whether to return the hamiltonian in sparse form.
    stype : str, optional
        What format of sparse operator to return if ``sparse``.
    parallel : bool, optional
        Whether to build the operator in parallel. By default will do this
        for n > 16.
    nthreads : int optional
        How mny threads to use in parallel to build the operator.
    ownership : (int, int), optional
        If given, which range of rows to generate.
    kwargs
        Supplied to :func:`~quimb.core.quimbify`.

    Returns
    -------
    H : immutable operator
        The Hamiltonian.
    """
    dims = (2,) * n
    try:
        jx, jy, jz = j
    except TypeError:
        jx = jy = jz = j

    try:
        bx, by, bz = b
    except TypeError:
        bz = b
        bx = by = 0.0

    parallel = (n > 16) if parallel is None else parallel

    op_kws = {'sparse': True, 'stype': 'coo'}
    ikron_kws = {'sparse': True, 'stype': 'coo',
                 'coo_build': True, 'ownership': ownership}

    # The basic operator (interaction and single b-field) that can be repeated.
    two_site_term = sum(
        j * kron(spin_operator(s, **op_kws), spin_operator(s, **op_kws))
        for j, s in zip((jx, jy, jz), 'xyz')
    ) - sum(
        b * kron(spin_operator(s, **op_kws), eye(2, **op_kws))
        for b, s in zip((bx, by, bz), 'xyz') if b != 0.0
    )

    single_site_b = sum(-b * spin_operator(s, **op_kws)
                        for b, s in zip((bx, by, bz), 'xyz') if b != 0.0)

    def gen_term(i):
        # special case: the last b term needs to be added manually
        if i == -1:
            return ikron(single_site_b, dims, n - 1, **ikron_kws)

        # special case: the interaction between first and last spins if cyclic
        if i == n - 1:
            return sum(
                j * ikron(spin_operator(s, **op_kws),
                          dims, [0, n - 1], **ikron_kws)
                for j, s in zip((jx, jy, jz), 'xyz') if j != 0.0)

        # General term, on-site b-field plus interaction with next site
        return ikron(two_site_term, dims, [i, i + 1], **ikron_kws)

    terms_needed = range(0 if not any((bx, by, bz)) else -1,
                         n if cyclic else n - 1)

    if parallel:
        pool = get_thread_pool(nthreads)
        ham = par_reduce(add, pool.map(gen_term, terms_needed))
    else:
        ham = sum(map(gen_term, terms_needed))

    return ham


def ham_ising(n, jz=1.0, bx=1.0, **ham_opts):
    """Generate the quantum transverse field ising model hamiltonian. This is a
    simple alias for :func:`~quimb.gen.operators.ham_heis` with Z-interactions
    and an X-field.
    """
    return ham_heis(n, j=(0, 0, jz), b=(bx, 0, 0), **ham_opts)


def ham_XY(n, jxy, bz, **ham_opts):
    """Generate the quantum transverse field XY model hamiltonian. This is a
    simple alias for :func:`~quimb.gen.operators.ham_heis` with
    X- and Y-interactions and a Z-field.
    """
    return ham_heis(n, j=(jxy, jxy, 0), b=(0, 0, bz), **ham_opts)


def ham_XXZ(n, delta, jxy=1.0, **ham_opts):
    """Generate the XXZ-model hamiltonian. This is a
    simple alias for :func:`~quimb.gen.operators.ham_heis` with matched
    X- and Y-interactions and ``delta`` Z coupling.
    """
    return ham_heis(n, j=(jxy, jxy, delta), b=0, **ham_opts)


@functools.lru_cache(maxsize=8)
@hamiltonian_builder
def ham_j1j2(n, j1=1.0, j2=0.5, bz=0.0, cyclic=True, ownership=None):
    """Generate the j1-j2 hamiltonian, i.e. next nearest neighbour
    interactions.

    Parameters
    ----------
    n : int
        Number of spins.
    j1 : float, optional
        Nearest neighbour coupling strength.
    j2 : float, optional
        Next nearest neighbour coupling strength.
    bz : float, optional
        B-field strength in z-direction.
    cyclic : bool, optional
        Cyclic boundary conditions.
    sparse : bool, optional
        Return hamiltonian as sparse-csr operator.
    ownership : (int, int), optional
        If given, which range of rows to generate.
    kwargs
        Supplied to :func:`~quimb.core.quimbify`.

    Returns
    -------
    H : immutable operator
        The Hamiltonian.
    """
    dims = (2,) * n

    op_kws = {'sparse': True, 'stype': 'coo'}
    ikron_kws = {'sparse': True, 'stype': 'coo',
                 'coo_build': True, 'ownership': ownership}

    sxyz = [spin_operator(i, **op_kws) for i in 'xyz']

    coosj1 = np.array([(i, i + 1) for i in range(n)])
    coosj2 = np.array([(i, i + 2) for i in range(n)])
    if cyclic:
        coosj1, coosj2 = coosj1 % n, coosj2 % n
    else:
        coosj1 = coosj1[np.all(coosj1 < n, axis=1)]
        coosj2 = coosj2[np.all(coosj2 < n, axis=1)]

    def j1_terms():
        for coo in coosj1:
            if abs(coo[1] - coo[0]) == 1:  # can sum then tensor (faster)
                yield ikron(sum(op & op for op in sxyz),
                            dims, coo, **ikron_kws)
            else:  # tensor then sum (slower)
                yield sum(ikron(op, dims, coo, **ikron_kws) for op in sxyz)

    def j2_terms():
        for coo in coosj2:
            if abs(coo[1] - coo[0]) == 2:  # can add then tensor (faster)
                yield ikron(sum(op & eye(2, **op_kws) & op for op in sxyz),
                            dims, coo, **ikron_kws)
            else:
                yield sum(ikron(op, dims, coo, **ikron_kws) for op in sxyz)

    ham = j1 * sum(j1_terms()) + j2 * sum(j2_terms())

    if bz != 0:
        gen_bz = (ikron([sxyz[2]], dims, i, **ikron_kws) for i in range(n))
        ham += bz * sum(gen_bz)

    return ham


def _gen_mbl_random_factors(n, dh, dh_dim, dh_dist, seed=None, beta=None):
    # sort out a vector of noise strengths -> e.g. (0, 0, 1) for z-noise only
    if isinstance(dh, (tuple, list)):
        dhds = dh
    else:
        dh_dim = {0: '', 1: 'z', 2: 'xy', 3: 'xyz'}.get(dh_dim, dh_dim)
        dhds = tuple((dh if d in dh_dim else 0) for d in 'xyz')

    if seed is not None:
        np.random.seed(seed)

    # sort out the noise distribution
    if dh_dist in {'g', 'gauss', 'gaussian', 'normal'}:
        rs = np.random.randn(3, n)

    elif dh_dist in {'s', 'flat', 'square', 'uniform', 'box'}:
        rs = 2.0 * np.random.rand(3, n) - 1.0

    elif dh_dist in {'qp', 'quasiperiodic', 'qr', 'quasirandom'}:
        if dh_dim != 'z':
            raise ValueError("dh_dim should be 1 or 'z' for dh_dist='qp'.")

        if beta is None:
            beta = (5**0.5 - 1) / 2

        # the random phase
        delta = 2 * np.pi * np.random.rand()

        # make sure get 3 by n different strengths
        inds = np.broadcast_to(range(n), (3, n))

        rs = np.cos(2 * np.pi * beta * inds + delta)

    return dhds, rs


@hamiltonian_builder
def ham_mbl(n, dh, j=1.0, bz=0.0, cyclic=True,
            seed=None, dh_dist="s", dh_dim=1, beta=None, ownership=None):
    """ Constructs a heisenberg hamiltonian with isotropic coupling and
    random fields acting on each spin - the many-body localized (MBL)
    spin hamiltonian.

    Parameters
    ----------
    n : int
        Number of spins.
    dh : float or (float, float, float)
        Strength of random fields (stdev of gaussian distribution), can be
        scalar (isotropic noise) or 3-vector for (x, y, z) directions.
    j : float or (float, float, float), optional
        Coupling strength, can be scalar (isotropic) or 3-vector.
    bz : float, optional
        Global magnetic field (in z-direction).
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    seed : int, optional
        Number to seed random number generator with.
    dh_dist : {'g', 's', 'qr'}, optional
        Type of random distribution for the noise:

        - "s": square, with bounds ``(-dh, dh)``
        - "g": gaussian, with standard deviation ``dh``
        - "qp": quasi periodic, with amplitude ``dh`` and
          'wavenumber' ``beta`` so that the field at site ``i`` is
          ``dh * cos(2 * pi * beta * i + delta)`` with ``delta`` a random
          offset between ``(0, 2 * pi)``, possibly seeded by ``seed``.

    dh_dim : {1, 2, 3} or str, optional
        The number of dimensions the noise acts in, or string
        specifier like ``'yz'``.
    beta : float, optional
        The wave number if ``dh_dist='qr'``, defaults to the golden
        ratio``(5**0.5 - 1) / 2``.
    sparse : bool, optional
        Whether to construct the hamiltonian in sparse form.
    stype : {'csr', 'csc', 'coo'}, optional
        The sparse format.
    ownership : (int, int), optional
        If given, which range of rows to generate.
    kwargs
        Supplied to :func:`~quimb.core.quimbify`.

    Returns
    -------
    H : operator
        The MBL hamiltonian for spin-1/2.

    See Also
    --------
    MPO_ham_mbl
    """
    dhds, rs = _gen_mbl_random_factors(n, dh, dh_dim, dh_dist, seed, beta)

    # the base hamiltonian ('csr' is most efficient format to add with)
    ham = ham_heis(n=n, j=j, b=bz, cyclic=cyclic,
                   sparse=True, stype='csr', ownership=ownership)

    op_kws = {'sparse': True, 'stype': 'coo'}
    ikron_kws = {'sparse': True, 'stype': 'coo',
                 'coo_build': True, 'ownership': ownership}

    def dh_terms():
        for i in range(n):
            # dhd - the total strength in direction x, y, or z
            # r - the random strength in direction x, y, or z for site i
            hdh = sum(dhd * r * spin_operator(s, **op_kws)
                      for dhd, r, s in zip(dhds, rs[:, i], 'xyz'))
            yield ikron(hdh, (2,) * n, i, **ikron_kws)

    ham = ham + sum(dh_terms())

    return ham


@hamiltonian_builder
def ham_heis_2D(n, m, j=1.0, bz=0.0, cyclic=False,
                parallel=False, ownership=None):
    """Construct the 2D spin-1/2 heisenberg model hamiltonian.

    Parameters
    ----------
    n : int
        The number of rows.
    m : int
        The number of columns.
    j : float or (float, float, float), optional
        The coupling strength(s). Isotropic if scalar else if
        vector ``(Jx, Jy, Jz) = j``.
    bz : float, optional
        The z direction magnetic field.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    sparse : bool, optional
        Whether to construct the hamiltonian in sparse form.
    stype : {'csr', 'csc', 'coo'}, optional
        The sparse format.
    parallel : bool, optional
        Construct the hamiltonian in parallel. Faster but might use more
        memory.
    ownership : (int, int), optional
        If given, which range of rows to generate.
    kwargs
        Supplied to :func:`~quimb.core.quimbify`.

    Returns
    -------
    H : operator
        The hamiltonian.
    """

    # parse interaction strengths
    try:
        jx, jy, jz = j
    except (TypeError, ValueError):
        jx = jy = jz = j

    dims = [[2] * m] * n  # shape (n, m)

    sites = tuple(itertools.product(range(n), range(m)))

    # generate neighbouring pair coordinates
    def gen_pairs():
        for i, j in sites:
            above, right = (i + 1) % n, (j + 1) % m
            # ignore wraparound coordinates if not cyclic
            if cyclic or above != 0:
                yield ((i, j), (above, j))
            if cyclic or right != 0:
                yield ((i, j), (i, right))

    # build the hamiltonian in sparse 'coo' format always for efficiency
    op_kws = {'sparse': True, 'stype': 'coo'}
    ikron_kws = {'sparse': True, 'stype': 'coo',
                 'coo_build': True, 'ownership': ownership}

    # generate all pairs of coordinates and directions
    pairs_ss = tuple(itertools.product(gen_pairs(), 'xyz'))

    # generate XX, YY and ZZ interaction from
    #     e.g. arg ([(3, 4), (3, 5)], 'z')
    def interactions(pair_s):
        pair, s = pair_s
        Sxyz = spin_operator(s, **op_kws)
        J = {'x': jx, 'y': jy, 'z': jz}[s]
        return ikron(J * Sxyz, dims, inds=pair, **ikron_kws)

    # generate Z field
    def fields(site):
        Sz = spin_operator('z', **op_kws)
        return ikron(bz * Sz, dims, inds=[site], **ikron_kws)

    if not parallel:
        # combine all terms
        all_terms = itertools.chain(
            map(interactions, pairs_ss),
            map(fields, sites) if bz != 0.0 else ())
        H = sum(all_terms)
    else:
        pool = get_thread_pool()
        all_terms = itertools.chain(
            pool.map(interactions, pairs_ss),
            pool.map(fields, sites) if bz != 0.0 else ())
        H = par_reduce(add, all_terms)

    return H


def uniq_perms(xs):
    """Generate all the unique permutations of sequence ``xs``.

    Examples
    --------
    >>> list(uniq_perms('0011'))
    [('0', '0', '1', '1'),
     ('0', '1', '0', '1'),
     ('0', '1', '1', '0'),
     ('1', '0', '0', '1'),
     ('1', '0', '1', '0'),
     ('1', '1', '0', '0')]
    """
    if len(xs) == 1:
        yield (xs[0],)
    else:
        uniq_xs = unique(xs)
        for first_x in uniq_xs:
            rem_xs = list(xs)
            rem_xs.remove(first_x)
            for sub_perm in uniq_perms(rem_xs):
                yield (first_x,) + sub_perm


@functools.lru_cache(maxsize=8)
def zspin_projector(n, sz=0, stype="csr", dtype=float):
    """Construct the projector onto spin-z subpspaces.

    Parameters
    ----------
    n : int
        Total size of spin system.
    sz : float or sequence of floats
        Spin-z value(s) subspace(s) to find projector for.
    stype : str
        Sparse format of the output operator.
    dtype : {float, complex}, optional
        The data type of the operator to generate.

    Returns
    -------
    prj : immutable sparse operator, shape (2**n, D)
        The (non-square) projector onto the specified subspace(s). The subspace
        size ``D`` is given by ``n choose (n / 2 + s)`` for each ``s``
        specified in ``sz``.

    Examples
    --------
    >>> zspin_projector(n=2, sz=0).A
    array([[0., 0.],
           [1., 0.],
           [0., 1.],
           [0., 0.]]

    Project a 9-spin Heisenberg-Hamiltonian into its spin-1/2 subspace:

    >>> H = ham_heis(9, sparse=True)
    >>> H.shape
    (512, 512)

    >>> P = zspin_projector(n=9, sz=1 / 2)
    >>> H0 = P.T @ H @ P
    >>> H0.shape
    (126, 126)
    """
    if not isiterable(sz):
        sz = (sz,)

    p = 0
    all_perms = []

    for s in sz:
        # Number of 'up' spins
        k = n / 2 + s
        if not k.is_integer():
            raise ValueError(f"{s} is not a valid spin half subspace for {n} "
                             "spins.")
        k = int(round(k))
        # Size of subspace
        p += comb(n, k, exact=True)
        # Find all computational basis states with correct number of 0s and 1s
        base_perm = '0' * (n - k) + '1' * k
        all_perms += [uniq_perms(base_perm)]

    # Coordinates
    cis = tuple(range(p))  # arbitrary basis
    cjs = tuple(int("".join(perm), 2) for perm in concat(all_perms))

    # Construct matrix which projects only on to these basis states
    prj = sp.coo_matrix((np.ones(p, dtype=dtype), (cjs, cis)),
                        shape=(2**n, p), dtype=dtype)
    prj = qu(prj, stype=stype, dtype=dtype)
    make_immutable(prj)
    return prj


@functools.lru_cache(8)
def create(n=2, **qu_opts):
    """The creation operator acting on an n-level system.
    """
    data = np.zeros((n, n))

    for i in range(n):
        data[i, i - 1] = i ** 0.5

    ap = qu(data, **qu_opts)
    make_immutable(ap)
    return ap


@functools.lru_cache(8)
def destroy(n=2, **qu_opts):
    """The annihilation operator acting on an n-level system.
    """
    am = create(n, **qu_opts).T.copy()
    make_immutable(am)
    return am


@functools.lru_cache(8)
def num(n, **qu_opts):
    """The number operator acting on an n-level system.
    """
    ap, am = create(n, **qu_opts), destroy(n, **qu_opts)
    an = qu(ap @ am, **qu_opts)
    make_immutable(an)
    return an


@functools.lru_cache(maxsize=8)
@hamiltonian_builder
def ham_hubbard_hardcore(n, t=0.5, V=1., mu=1., cyclic=True,
                         parallel=False, ownership=None):
    """Generate the spinless fermion hopping hamiltonian.

    Parameters
    ----------
    n : int
        The number of sites.
    t : float, optional
        The hopping energy.
    V : float, optional
        The interaction energy.
    mu : float, optional
        The chemical potential - defaults to half-filling.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    parallel : bool, optional
        Construct the hamiltonian in parallel. Faster but might use more
        memory.
    ownership : (int, int), optional
        If given, which range of rows to generate.
    kwargs
        Supplied to :func:`~quimb.core.quimbify`.

    Returns
    -------
    H : operator
        The hamiltonian.
    """

    op_kws = {'sparse': True, 'stype': 'coo'}
    ikron_kws = {'sparse': True, 'stype': 'csr',
                 'coo_build': True, 'ownership': ownership}

    cdag, c, cnum = (f(2, **op_kws) for f in (create, destroy, num))
    neighbor_term = t * ((cdag & c) + (c & cdag)) + V * (cnum & cnum)

    dims = [2] * n

    def terms():
        # interacting terms
        for i, j in [(i, i + 1) for i in range(n - 1)]:
            yield ikron(neighbor_term, dims, (i, j), **ikron_kws)

        if cyclic:
            # can't sum terms and kron later since identity in middle
            yield ikron([t * cdag, c], dims, (0, n - 1), **ikron_kws)
            yield ikron([t * c, cdag], dims, (0, n - 1), **ikron_kws)
            yield ikron([V * cnum, cnum], dims, (0, n - 1), **ikron_kws)

        # single site terms
        for i in range(n):
            yield ikron(-mu * cnum, dims, i, **ikron_kws)

    if parallel is None:
        parallel = (n >= 14)

    if parallel:
        return par_reduce(operator.add, terms())

    return functools.reduce(operator.add, terms())



import numpy as np
import numba

from ..core import njit, pnjit, qarray


@njit
def get_nz(A):  # pragma: no cover
    return np.nonzero(A)


@njit(['void(int32, int32, List(Set(int32)))',
       'void(int64, int64, List(Set(int64)))'])
def _add_to_groups(i, j, groups):  # pragma: no cover
    for group in groups:
        if i in group:
            group.add(j)
            return
        if j in group:
            group.add(i)
            return
    # pair is not in a sector yet - create new one
    groups.append({i, j})


@njit(['List(List(int32))(int32[:], int32[:], int_)',
       'List(List(int64))(int64[:], int64[:], int_)'])
def compute_blocks(ix, jx, d):  # pragma: no cover
    """Find the charge sectors (blocks in matrix terms) given element
    coordinates ``ix`` and ``jx`` and total size ``d``.

    Parameters
    ----------
    ix : array of int
        The row coordinates of non-zero elements.
    jx : array of int
        The column coordinates of non-zero elements.
    d : int
        The total size of the operator.

    Returns
    -------
    sectors : list[list[int]]
        The list of charge sectors. Each element is itself a sorted list of the
        basis numbers that make up that sector. The permutation that would
        block diagonalize the operator is then ``np.concatenate(sectors)``.

    Examples
    --------

    >>> H = ham_hubbard_hardcore(4, sparse=True)
    >>> ix, jx = H.nonzero()
    >>> d = H.shape[0]
    >>> sectors = compute_blocks(ix, jx, d)
    >>> sectors
    [[0], [1, 2, 4, 8], [3, 5, 6, 9, 10, 12], [7, 11, 13, 14], [15]]
    """
    groups = [{ix[0], jx[0]}]

    # go through actual nz
    for i, j in zip(ix, jx):
        _add_to_groups(i, j, groups)

    # make sure kernel added as subspace
    for i in range(d):
        _add_to_groups(i, i, groups)

    # sort indices in each group and groups by first element
    return sorted([sorted(g) for g in groups])


@pnjit
def subselect(A, p):  # pragma: no cover
    """Select only the intersection of rows and columns of ``A`` matching the
    basis indices ``p``. Faster than double numpy slicing.

    Parameters
    ----------
    A : 2D-array
        Dense matrix to select from.
    p : sequence of int
        The basis indices to select.

    Returns
    -------
    B : 2D-array
        The matrix, of size ``(len(p), len(p))``.

    Examples
    --------
    >>> A = np.arange(25).reshape(5, 5)
    >>> A
    array([[ 0,  1,  2,  3,  4],
           [ 5,  6,  7,  8,  9],
           [10, 11, 12, 13, 14],
           [15, 16, 17, 18, 19],
           [20, 21, 22, 23, 24]])

    >>> subselect(A, [1, 3])
    array([[ 6,  8],
           [16, 18]])
    """
    dp = len(p)
    out = np.empty((dp, dp), dtype=A.dtype)

    for i in numba.prange(dp):
        for j in numba.prange(dp):
            out[i, j] = A[p[i], p[j]]

    return out


@pnjit
def subselect_set(A, B, p):  # pragma: no cover
    """Set only the intersection of rows and colums of ``A`` matching the
    basis indices ``p`` to ``B``.

    Parameters
    ----------
    A : array with shape (d, d)
        The matrix to set elements in.
    B : array with shape (dp, dp)
        The matrix to set elements from.
    p : sequence of size dp
        The basis indices.

    Examples
    --------
    >>> A = np.zeros((5, 5))
    >>> B = np.random.randn(3, 3)
    >>> p = [0, 2, 4]
    >>> subselect_set(A, B, p)
    array([[-0.31888218,  0.        ,  0.39293245,  0.        ,  0.21822712],
           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ],
           [ 0.66674486,  0.        ,  1.03388035,  0.        ,  1.7319345 ],
           [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ],
           [-0.94542733,  0.        , -0.37211882,  0.        ,  0.51951555]])
    """
    dp = len(p)

    for i in numba.prange(dp):
        for j in numba.prange(dp):
            A[p[i], p[j]] = B[i, j]


# XXX: want to cache this eventaully -> need parallel+cache numba support?
@njit
def _eigh_autoblocked(A, sort=True):  # pragma: no cover
    d = A.shape[0]

    # allocate output arrays
    el = np.empty(d)
    ev = np.zeros_like(A)

    # find non-zero elements and group into charge sectors
    ix, jx = get_nz(A)
    gs = compute_blocks(ix, jx, d)
    gs = [np.array(g) for g in gs]

    # diagonalize each charge sector seperately
    for i, g in enumerate(gs):
        ng = len(g)

        # check if trivial
        if ng == 1:
            el[g[0]] = A[g[0], g[0]].real
            ev[g[0], g[0]] = 1.0
            continue

        # else diagonalize just the block
        sub_el, sub_ev = np.linalg.eigh(subselect(A, g))

        # set the correct eigenpairs in the output
        el[g] = sub_el
        subselect_set(ev, sub_ev, g)

    # sort into ascending eigenvalue order
    if sort:
        so = np.argsort(el)
        el[:] = el[so]
        ev[:, :] = ev[:, so]

    return el, ev


# XXX: want to cache this eventaully -> need parallel+cache numba support?
@njit
def _eigvalsh_autoblocked(A, sort=True):  # pragma: no cover
    # as above but ignore eigenvector for extra speed
    d = A.shape[0]

    el = np.empty(d)

    ix, jx = get_nz(A)
    gs = compute_blocks(ix, jx, d)
    gs = [np.array(g) for g in gs]

    for i, g in enumerate(gs):
        if len(g) == 1:
            el[g[0]] = A[g[0], g[0]]
            continue

        el[g] = np.linalg.eigvalsh(subselect(A, g))

    if sort:
        el[:] = np.sort(el)

    return el


def eigensystem_autoblocked(A, sort=True, return_vecs=True, isherm=True):
    """Perform Hermitian eigen-decomposition, automatically identifying and
    exploiting symmetries appearing in the current basis as block diagonals
    formed via permutation of rows and columns. The whole process is
    accelerated using ``numba``.

    Parameters
    ----------
    A : array_like
        The operator to eigen-decompose.
    sort : bool, optional
        Whether to sort into ascending order, default True.
    isherm : bool, optional
        Whether ``A`` is hermitian, default True.
    return_vecs : bool, optional
        Whether to return the eigenvectors, default True.

    Returns
    -------
    evals : 1D-array
        The eigenvalues.
    evecs : qarray
        If ``return_vecs=True``, the eigenvectors.
    """
    if not isherm:
        err_msg = "Non-hermitian autoblocking not implemented yet."
        raise NotImplementedError(err_msg)

    if not return_vecs:
        return _eigvalsh_autoblocked(A, sort=sort)

    el, ev = _eigh_autoblocked(A, sort=sort)
    return el, qarray(ev)
